#    Qlattice (https://github.com/jinluchang/qlattice)
#
#    Copyright (C) 2021
#
#    Author: Luchang Jin (ljin.luchang@gmail.com)
#    Author: Masaaki Tomii
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import qlat_utils as q
import copy
import sympy
import numpy as np
from typing import Self

try:
    from . import expr_arithmetic as ea
    from .expr_arithmetic import mk_sym
except:
    import expr_arithmetic as ea
    from expr_arithmetic import mk_sym

class Op:

    """
    self.otype
    """

    def __init__(self, otype:str):
        self.otype = otype

    def is_commute(self) -> bool:
        return True

    def show(self, is_multiply=False):
        return repr(self)

    def __repr__(self) -> str:
        return f"Op({self.otype})"

    def __add__(self, other):
        return mk_expr(self) + other

    __radd__ = __add__

    def __mul__(self, other):
        return mk_expr(self) * other

    def __rmul__(self, other):
        return mk_expr(other) * self

    def __neg__(self):
        return mk_expr(-1) * mk_expr(self)

    def __pos__(self):
        return self

    def __sub__(self, other):
        return mk_expr(self) + mk_expr(-1) * other

    def __rsub__(self, other):
        return mk_expr(other) + mk_expr(-1) * self

    def sort(self) -> None:
        pass

    def isospin_symmetric_limit(self) -> None:
        pass

### ------

class Qfield(Op):

    """
    self.f
    self.p
    self.s
    self.c
    """

    def __init__(self, otype:str, flavor:str, position:str, spin:str, color:str):
        Op.__init__(self, otype)
        self.f = flavor
        self.p = position
        self.s = spin
        self.c = color

    def is_commute(self) -> bool:
        return False

    def __repr__(self) -> str:
        if self.s == "auto" and self.c == "auto":
            return f"{self.otype}({self.f!r},{self.p!r})"
        else:
            return f"{self.otype}({self.f!r},{self.p!r},{self.s!r},{self.c!r})"

    def list(self):
        return [ self.otype, self.f, self.p, self.s, self.c, ]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

### ------

class Qv(Qfield):

    """
    act as d / d Hb
    """

    def __init__(self, f, p, s, c):
        """
        interface function
        """
        Qfield.__init__(self, "Qv", f, p, s, c)

### ------

class Qb(Qfield):

    """
    act as d / d Hv
    """

    def __init__(self, f, p, s, c):
        """
        interface function
        """
        Qfield.__init__(self, "Qb", f, p, s, c)

### ------

class Hv(Qfield):

    def __init__(self, f, p, s, c):
        Qfield.__init__(self, "Hv", f, p, s, c)

### ------

class Hb(Qfield):

    def __init__(self, f, p, s, c):
        Qfield.__init__(self, "Hb", f, p, s, c)

### ------

class SHv(Qfield):

    def __init__(self, f, p, s, c):
        Qfield.__init__(self, "SHv", f, p, s, c)

### ------

class HbS(Qfield):

    def __init__(self, f, p, s, c):
        Qfield.__init__(self, "HbS", f, p, s, c)

### ------

class S(Op):

    """
    propagator
    #
    self.f
    self.p1
    self.p2
    self.s1
    self.s2
    self.c1
    self.c2
    """

    def __init__(self, flavor:str, p1:str, p2:str, s1:str="auto", s2:str="auto", c1:str="auto", c2:str="auto"):
        Op.__init__(self, "S")
        self.f = flavor
        self.p1 = p1
        self.p2 = p2
        self.s1 = s1
        self.s2 = s2
        self.c1 = c1
        self.c2 = c2

    def __repr__(self) -> str:
        if self.s1 == "auto" and self.s2 == "auto" and self.c1 == "auto" and self.c2 == "auto":
            return f"{self.otype}({self.f!r},{self.p1!r},{self.p2!r})"
        else:
            return f"{self.otype}({self.f!r},{self.p1!r},{self.p2!r},{self.s1!r},{self.s2!r},{self.c1!r},{self.c2!r})"

    def list(self):
        return [ self.otype, self.f, self.p1, self.p2, self.s1, self.s2, self.c1, self.c2, ]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

    def isospin_symmetric_limit(self) -> None:
        """
        can also use these fictitious quark field to remove some unwanted disconnected diagrams
        """
        if self.f in [ "u", "d", "u'", "d'", "u''", "d''", "u'''", "d'''", "u''''", "d''''", ]:
            self.f = "l"
        elif self.f in [ "l", "l'", "l''", "l'''", "l''''", ]:
            self.f = "l"
        elif self.f in [ "s", "s'", "s''", "s'''", "s''''", ]:
            self.f = "s"
        elif self.f in [ "c", "c'", "c''", "c'''", "c''''", ]:
            self.f = "c"
        elif self.f in [ "b", "b'", "b''", "b'''", "b''''", ]:
            self.f = "b"
        elif self.f in [ "t", "t'", "t''", "t'''", "t''''", ]:
            self.f = "t"
        else:
            fname = q.get_fname()
            raise Exception(f"{fname}: self.f=\"{self.f}\"")

### ------

class G(Op):

    """
    spin matrix
    #
    self.tag
    self.s1
    self.s2
    #
    tag = 0, 1, 2, 3, 5 for gamma matrices
    """

    def __init__(self, tag, s1:str="auto", s2:str="auto"):
        Op.__init__(self, "G")
        self.tag = tag
        self.s1 = s1
        self.s2 = s2

    def __repr__(self) -> str:
        if self.s1 == "auto" and self.s2 == "auto":
            return f"{self.otype}({self.tag!r})"
        else:
            return f"{self.otype}({self.tag!r},{self.s1!r},{self.s2!r})"

    def list(self):
        return [ self.otype, self.tag, self.s1, self.s2, ]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

### ------

class U(Op):

    """
    color matrix
    #
    self.tag
    self.p
    self.mu
    self.c1
    self.c2
    """

    def __init__(self, tag, p, mu, c1:str="auto", c2:str="auto"):
        Op.__init__(self, "U")
        self.tag = tag
        self.p = p
        self.mu = mu
        self.c1 = c1
        self.c2 = c2

    def __repr__(self) -> str:
        if self.c1 == "auto" and self.c2 == "auto":
            return f"{self.otype}({self.tag!r},{self.p!r},{self.mu!r})"
        else:
            return f"{self.otype}({self.tag!r},{self.p!r},{self.mu!r},{self.c1!r},{self.c2!r})"

    def list(self):
        return [ self.otype, self.tag, self.p, self.mu, self.c1, self.c2, ]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

### ------

class Tr(Op):

    """
    a collection of ops taking the trace
    #
    self.ops
    self.tag
    """

    def __init__(self, ops:list, tag=None):
        Op.__init__(self, "Tr")
        if tag is not None:
            # do not perform check if tag is set
            self.ops = ops
            self.tag = tag
            return
        for op in ops:
            assert op.is_commute()
            assert op.otype in [ "S", "G", "U", ]
        s = None
        c = None
        for op in ops + ops:
            if not check_chain_sc(op, s, c):
                raise Exception(f"ops={ops} tag={tag}")
            s, c, = update_chain_sc(op, s, c)
        if s is not None and c is not None:
            self.tag = "sc"
        elif s is not None:
            self.tag = "s"
        elif c is not None:
            self.tag = "c"
        else:
            self.tag = ""
        self.ops = []
        for op in ops:
            self.ops.append(copy_op_index_auto(op))

    def __repr__(self) -> str:
        return f"{self.otype}({self.ops!r},{self.tag!r})"

    def list(self):
        return [ self.otype, self.tag, self.ops, ]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

    def sort(self):
        ops = self.ops
        if len(ops) > 1:
            self.ops = sorted([ ops[i:] + ops[:i] for i in range(len(ops)) ], key=repr)[0]

    def isospin_symmetric_limit(self) -> None:
        for op in self.ops:
            op.isospin_symmetric_limit()

### ------

class Chain(Op):

    """
    a collection of ops multiplying together but do not form a loop
    #
    self.ops
    self.tag
    self.s1
    self.s2
    self.c1
    self.c2
    """

    def __init__(self, ops:list, tag=None, s1:str="auto", s2:str="auto", c1:str="auto", c2:str="auto"):
        Op.__init__(self, "Chain")
        if tag is not None:
            # do not perform check if tag is set
            self.ops = ops
            self.s1 = s1
            self.s2 = s2
            self.c1 = c1
            self.c2 = c2
            self.tag = tag
            return
        for op in ops:
            assert op.is_commute()
            assert op.otype in [ "S", "G", "U", ]
        self.s1 = None
        self.s2 = None
        self.c1 = None
        self.c2 = None
        s = None
        c = None
        for op in ops:
            if not check_chain_sc(op, s, c):
                raise Exception(f"ops={ops} tag={tag}")
            s, c, = update_chain_sc1(op, s, c)
            if self.s1 is None:
                self.s1 = s
            if self.c1 is None:
                self.c1 = c
            s, c, = update_chain_sc(op, s, c)
        self.s2 = s
        self.c2 = c
        if s is not None and c is not None:
            self.tag = "sc"
        elif s is not None:
            self.tag = "s"
        elif c is not None:
            self.tag = "c"
        else:
            self.tag = ""
        self.ops = []
        for op in ops:
            self.ops.append(copy_op_index_auto(op))

    def __repr__(self) -> str:
        if self.s1 == "auto" and self.s2 == "auto" and self.c1 == "auto" and self.c2 == "auto":
            return f"{self.otype}({self.ops!r},{self.tag!r})"
        else:
            return f"{self.otype}({self.ops!r},{self.tag!r},{self.s1!r},{self.s2!r},{self.c1!r},{self.c2!r})"

    def list(self):
        return [ self.otype, self.ops, self.tag, self.s1, self.s2, self.c1, self.c2, ]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

    def sort(self):
        """
        Do not need to sort
        """

    def isospin_symmetric_limit(self) -> None:
        for op in self.ops:
            op.isospin_symmetric_limit()

### ------

def copy_op_index_auto(op:Op):
    if op.otype not in [ "S", "G", "U", "Chain", ]:
        return op
    op = copy.copy(op)
    if op.otype in [ "S", "Chain", "G", ]:
        op.s1 = "auto"
        op.s2 = "auto"
    if op.otype in [ "S", "Chain", "U", ]:
        op.c1 = "auto"
        op.c2 = "auto"
    return op

def check_chain_spin_index(ops:list, s:str):
    """
    for a spin index `s`,
    return is_having_repeated_index, first_index_of_operator, second_index_of_operator
    """
    count1 = 0
    count2 = 0
    i1 = None
    i2 = None
    for i, op in enumerate(ops):
        if op.otype in [ "S", "G", ]:
            if op.s1 == s:
                i1 = i
                count1 += 1
            if op.s2 == s:
                i2 = i
                count2 += 1
    assert count1 < 2
    assert count2 < 2
    return count1 == 1 and count2 == 1, i1, i2

def check_chain_color_index(ops:list, c:str):
    """
    for a color index `s`,
    return is_having_repeated_index, first_index_of_operator, second_index_of_operator
    """
    count1 = 0
    count2 = 0
    i1 = None
    i2 = None
    for i, op in enumerate(ops):
        if op.otype in [ "S", "U", ]:
            if op.c1 == c:
                i1 = i
                count1 += 1
            if op.c2 == c:
                i2 = i
                count2 += 1
    assert count1 < 2
    assert count2 < 2
    return count1 == 1 and count2 == 1, i1, i2

def check_chain_op(ops:list, op:Op):
    """
    for a operator `op`,
    return "single" or "begin" or "middle" or "end",
    if this `op` is part of a (longer) chain of contraction
    if `op` does not have right type, return `None`.
    """
    if op.otype not in [ "S", "G", "U", ]:
        return None
    b_begin = True
    b_end = True
    if op.otype in [ "S", ]:
        if check_chain_spin_index(ops, op.s1)[0] and check_chain_color_index(ops, op.c1)[0]:
            b_begin = False
        if check_chain_spin_index(ops, op.s2)[0] and check_chain_color_index(ops, op.c2)[0]:
            b_end = False
    elif op.otype in [ "G", ]:
        if check_chain_spin_index(ops, op.s1)[0]:
            b_begin = False
        if check_chain_spin_index(ops, op.s2)[0]:
            b_end = False
    elif op.otype in [ "U", ]:
        if check_chain_color_index(ops, op.c1)[0]:
            b_begin = False
        if check_chain_color_index(ops, op.c2)[0]:
            b_end = False
    else:
        assert False
    if b_begin and b_end:
        return "single"
    elif b_begin:
        return "begin"
    elif b_end:
        return "end"
    else:
        return "middle"

def check_chain_sc(op, s, c):
    if op.otype in [ "S", "G", ]:
        if s is not None and s != op.s1:
            return False
    if op.otype in [ "S", "U", ]:
        if c is not None and c != op.c1:
            return False
    return True

def update_chain_sc1(op, s, c):
    if op.otype in [ "S", "Chain", "G", ]:
        s = op.s1
    if op.otype in [ "S", "Chain", "U", ]:
        c = op.c1
    return s, c

def update_chain_sc(op, s, c):
    if op.otype in [ "S", "Chain", "G", ]:
        s = op.s2
    if op.otype in [ "S", "Chain", "U", ]:
        c = op.c2
    return s, c

def pick_chain_op(ops:list, masks:list, s, c, type_list=None):
    if type_list is None:
        type_list = [ "single", "begin", "end", "middle", ]
    for i, op in enumerate(ops):
        if masks[i]:
            continue
        if not check_chain_sc(op, s, c):
            continue
        if check_chain_op(ops, op) not in type_list:
            continue
        return i, op
    return None

def find_trace(ops:list):
    """
    return None or (Tr(tr_ops), remaining_op_list,)
    """
    size = len(ops)
    for i, op in enumerate(ops):
        if check_chain_op(ops, op) not in [ "middle", ]:
            continue
        masks = [ False for op in ops ]
        tr_ops = []
        s = None
        c = None
        masks[i] = True
        tr_ops.append(op)
        s, c, = update_chain_sc(op, s, c)
        while True:
            p_op = pick_chain_op(ops, masks, s, c, [ "middle", ])
            if p_op is None:
                # trace found
                return Tr(tr_ops), [ op for i, op in enumerate(ops) if not masks[i] ]
            i2, op2 = p_op
            masks[i2] = True
            tr_ops.append(op2)
            s, c, = update_chain_sc(op2, s, c)
    return None

def find_chain(ops:list):
    """
    return None or (Chain(ch_ops), remaining_op_list,)
    """
    size = len(ops)
    for i, op in enumerate(ops):
        if check_chain_op(ops, op) not in [ "single", "begin", ]:
            continue
        masks = [ False for op in ops ]
        ch_ops = []
        s = None
        c = None
        masks[i] = True
        ch_ops.append(op)
        s, c, = update_chain_sc(op, s, c)
        while True:
            p_op = pick_chain_op(ops, masks, s, c, [ "single", "begin", "middle", "end", ])
            if p_op is None:
                # chain found
                return Chain(ch_ops), [ op for i, op in enumerate(ops) if not masks[i] ]
            i2, op2 = p_op
            masks[i2] = True
            ch_ops.append(op2)
            s, c, = update_chain_sc(op2, s, c)
    return None

@q.timer
def collect_traces(ops:list) -> list:
    """
    First collect all the `Chain`s, then `Tr`s
    """
    chs = []
    while True:
        fc = find_chain(ops)
        if fc is None:
            break
        ch, ops = fc
        chs.append(ch)
    trs = []
    while True:
        ft = find_trace(ops)
        if ft is None:
            break
        tr, ops = ft
        trs.append(tr)
    return chs + trs + ops

### ------

@q.timer
def baryon_spin_tensor_to_code(spin_tensor:np.ndarray) -> np.ndarray:
    """
    return spin_tensor_code
    spin_tensor_code[s1, s2, s3] = spin_coef
    Note CPS/Grid/GPT code convention is not the Euclidean convention
    #
    psi^Code_0 = -1 * psi^Eucl_1
    psi^Code_1 = +1 * psi^Eucl_0
    psi^Code_2 = -1 * psi^Eucl_3
    psi^Code_3 = +1 * psi^Eucl_2
    """
    shape = (4, 4, 4,)
    assert spin_tensor.shape == shape
    spin_tensor_code = np.zeros(shape, dtype=object)
    sidx_arr = [
            (1, -1),
            (0, 1),
            (3, -1),
            (2, 1),
            ]
    for s1 in range(4):
        s1p, fac1, = sidx_arr[s1]
        for s2 in range(4):
            s2p, fac2, = sidx_arr[s2]
            for s3 in range(4):
                s3p, fac3, = sidx_arr[s3]
                spin_tensor_code[s1, s2, s3] = fac1 * fac2 * fac3 * spin_tensor[s1p, s2p, s3p]
    return spin_tensor_code

@q.timer
def baryon_spin_tensor_from_code(spin_tensor_code:np.ndarray) -> np.ndarray:
    """
    return spin_tensor
    spin_tensor[s1, s2, s3] = spin_coef
    Note CPS/Grid/GPT code convention is not the Euclidean convention
    #
    psi^Eucl_0 = +1 * psi^Code_1
    psi^Eucl_1 = -1 * psi^Code_0
    psi^Eucl_2 = +1 * psi^Code_3
    psi^Eucl_3 = -1 * psi^Code_2
    """
    shape = (4, 4, 4,)
    assert spin_tensor_code.shape == shape
    spin_tensor = np.zeros(shape, dtype=object)
    sidx_arr = [
            (1, 1),
            (0, -1),
            (3, 1),
            (2, -1),
            ]
    for s1 in range(4):
        s1p, fac1, = sidx_arr[s1]
        for s2 in range(4):
            s2p, fac2, = sidx_arr[s2]
            for s3 in range(4):
                s3p, fac3, = sidx_arr[s3]
                spin_tensor[s1, s2, s3] = fac1 * fac2 * fac3 * spin_tensor_code[s1p, s2p, s3p]
    return spin_tensor

class BfieldCoef:

    """
    self.spin_tensor
    #
    baryon_operator =
    ( spin_tensor[spin_idx1, spin_idx2, spin_idx3]
    * epsilon[color_idx1, color_idx2, color_idx3]
    * quark_field_1[spin_idx1, color_idx1]
    * quark_field_2[spin_idx2, color_idx2]
    * quark_field_3[spin_idx3, color_idx3]
    )
    """

    def __init__(self, st_list_code:list|None=None):
        shape = (4, 4, 4,)
        if st_list_code is None:
            self.spin_tensor = np.zeros(shape, dtype=object)
        else:
            spin_tensor_code = np.zeros(shape, dtype=object)
            spin_tensor_code.ravel()[:] = st_list_code
            self.spin_tensor = baryon_spin_tensor_from_code(spin_tensor_code)

    def get_spin_tensor_list_code(self) -> list:
        st = self.get_spin_tensor_code()
        st_list_code = st.ravel().tolist()
        return st_list_code

    def get_spin_tensor(self, permute=None) -> np.ndarray:
        """
        return spin_tensor
        spin_tensor[s1, s2, s3] = spin_coef
        Euclidean convention, not the code convention
        """
        if permute is None:
            permute = (0, 1, 2,)
        spin_tensor = q.epsilon_tensor(*permute) * self.spin_tensor.transpose(permute)
        shape = (4, 4, 4,)
        assert spin_tensor.shape == shape
        return spin_tensor

    @q.timer
    def get_spin_tensor_code(self, permute=None) -> np.ndarray:
        """
        return spin_tensor_code
        See `baryon_spin_tensor_to_code` and `baryon_spin_tensor_from_code`.
        """
        spin_tensor = self.get_spin_tensor(permute=permute)
        spin_tensor_code = baryon_spin_tensor_to_code(spin_tensor)
        return spin_tensor_code

    def add(self, chiral_projection, spin, spin_coef) -> Self:
        """
        return self
        Allow keep add.
        #
        chiral_projection = ((1, 1,), (1, 0,), (1, 0,),)
        spin = (0, 1, 0,)
        spin_coef = 1
        """
        if spin_coef == 0:
            return self
        ((c00, c01,), (c10, c11,), (c20, c21,),) = chiral_projection
        s1, s2, s3, = spin
        spin_tensor = self.spin_tensor
        spin_tensor[s1, s2, s3] += c00 * c10 * c20 * spin_coef
        spin_tensor[s1, s2, s3 + 2] += c00 * c10 * c21 * spin_coef
        spin_tensor[s1, s2 + 2, s3] += c00 * c11 * c20 * spin_coef
        spin_tensor[s1, s2 + 2, s3 + 2] += c00 * c11 * c21 * spin_coef
        spin_tensor[s1 + 2, s2, s3] += c01 * c10 * c20 * spin_coef
        spin_tensor[s1 + 2, s2, s3 + 2] += c01 * c10 * c21 * spin_coef
        spin_tensor[s1 + 2, s2 + 2, s3] += c01 * c11 * c20 * spin_coef
        spin_tensor[s1 + 2, s2 + 2, s3 + 2] += c01 * c11 * c21 * spin_coef
        return self

    def list(self) -> list:
        return [ self.spin_tensor, ]

    def __repr__(self) -> str:
        st_list_code = self.get_spin_tensor_list_code()
        st_str = ",".join([ repr(v) for v in st_list_code ])
        return f"BfieldCoef([{st_str}])"

### ------

bfield_tag_dict = dict()

bfield_tag_dict["std-u"] = BfieldCoef()
bfield_tag_dict["std-u"].add(((1, 1,), (1, 0,), (1, 0,),), (0, 1, 0,), mk_sym(1)/2)
bfield_tag_dict["std-u"].add(((1, 1,), (1, 0,), (1, 0,),), (0, 0, 1,), -mk_sym(1)/2)
bfield_tag_dict["std-u"].add(((1, 1,), (0, 1,), (0, 1,),), (0, 1, 0,), mk_sym(1)/2)
bfield_tag_dict["std-u"].add(((1, 1,), (0, 1,), (0, 1,),), (0, 0, 1,), -mk_sym(1)/2)

bfield_tag_dict["std-d"] = BfieldCoef()
bfield_tag_dict["std-d"].add(((1, 1,), (1, 0,), (1, 0,),), (1, 1, 0,), mk_sym(1)/2)
bfield_tag_dict["std-d"].add(((1, 1,), (1, 0,), (1, 0,),), (1, 0, 1,), -mk_sym(1)/2)
bfield_tag_dict["std-d"].add(((1, 1,), (0, 1,), (0, 1,),), (1, 1, 0,), mk_sym(1)/2)
bfield_tag_dict["std-d"].add(((1, 1,), (0, 1,), (0, 1,),), (1, 0, 1,), -mk_sym(1)/2)

bfield_tag_dict["std3-u3"] = BfieldCoef()
bfield_tag_dict["std3-u3"].add(((1, 1,), (1, 0,), (0, 1,),), (0, 0, 0,), 1/sympy.sqrt(2))
bfield_tag_dict["std3-u3"].add(((1, 1,), (0, 1,), (1, 0,),), (0, 0, 0,), 1/sympy.sqrt(2))

bfield_tag_dict["std3-u1"] = BfieldCoef()
bfield_tag_dict["std3-u1"].add(((1, 1,), (1, 0,), (0, 1,),), (1, 0, 0,), 1/sympy.sqrt(6))
bfield_tag_dict["std3-u1"].add(((1, 1,), (0, 1,), (1, 0,),), (1, 0, 0,), 1/sympy.sqrt(6))
bfield_tag_dict["std3-u1"].add(((1, 1,), (1, 0,), (0, 1,),), (0, 1, 0,), 1/sympy.sqrt(6))
bfield_tag_dict["std3-u1"].add(((1, 1,), (0, 1,), (1, 0,),), (0, 1, 0,), 1/sympy.sqrt(6))
bfield_tag_dict["std3-u1"].add(((1, 1,), (1, 0,), (0, 1,),), (0, 0, 1,), 1/sympy.sqrt(6))
bfield_tag_dict["std3-u1"].add(((1, 1,), (0, 1,), (1, 0,),), (0, 0, 1,), 1/sympy.sqrt(6))

bfield_tag_dict["std3-d1"] = BfieldCoef()
bfield_tag_dict["std3-d1"].add(((1, 1,), (1, 0,), (0, 1,),), (1, 1, 0,), 1/sympy.sqrt(6))
bfield_tag_dict["std3-d1"].add(((1, 1,), (0, 1,), (1, 0,),), (1, 1, 0,), 1/sympy.sqrt(6))
bfield_tag_dict["std3-d1"].add(((1, 1,), (1, 0,), (0, 1,),), (0, 1, 1,), 1/sympy.sqrt(6))
bfield_tag_dict["std3-d1"].add(((1, 1,), (0, 1,), (1, 0,),), (0, 1, 1,), 1/sympy.sqrt(6))
bfield_tag_dict["std3-d1"].add(((1, 1,), (1, 0,), (0, 1,),), (1, 0, 1,), 1/sympy.sqrt(6))
bfield_tag_dict["std3-d1"].add(((1, 1,), (0, 1,), (1, 0,),), (1, 0, 1,), 1/sympy.sqrt(6))

bfield_tag_dict["std3-d3"] = BfieldCoef()
bfield_tag_dict["std3-d3"].add(((1, 1,), (1, 0,), (0, 1,),), (1, 1, 1,), 1/sympy.sqrt(2))
bfield_tag_dict["std3-d3"].add(((1, 1,), (0, 1,), (1, 0,),), (1, 1, 1,), 1/sympy.sqrt(2))

bfield_tag_dict["pos-u"] = BfieldCoef()
bfield_tag_dict["pos-u"].add(((1, 1,), (1, 1,), (1, 1,),), (0, 1, 0,), mk_sym(1)/2)
bfield_tag_dict["pos-u"].add(((1, 1,), (1, 1,), (1, 1,),), (0, 0, 1,), -mk_sym(1)/2)

bfield_tag_dict["pos-d"] = BfieldCoef()
bfield_tag_dict["pos-d"].add(((1, 1,), (1, 1,), (1, 1,),), (1, 1, 0,), mk_sym(1)/2)
bfield_tag_dict["pos-d"].add(((1, 1,), (1, 1,), (1, 1,),), (1, 0, 1,), -mk_sym(1)/2)

bfield_tag_dict["pos3-u3"] = BfieldCoef()
bfield_tag_dict["pos3-u3"].add(((1, 1,), (1, 1,), (1, 1,),), (0, 0, 0,), 1)

bfield_tag_dict["pos3-u1"] = BfieldCoef()
bfield_tag_dict["pos3-u1"].add(((1, 1,), (1, 1,), (1, 1,),), (1, 0, 0,), 1/sympy.sqrt(3))
bfield_tag_dict["pos3-u1"].add(((1, 1,), (1, 1,), (1, 1,),), (0, 1, 0,), 1/sympy.sqrt(3))
bfield_tag_dict["pos3-u1"].add(((1, 1,), (1, 1,), (1, 1,),), (0, 0, 1,), 1/sympy.sqrt(3))

bfield_tag_dict["pos3-d1"] = BfieldCoef()
bfield_tag_dict["pos3-d1"].add(((1, 1,), (1, 1,), (1, 1,),), (1, 1, 0,), 1/sympy.sqrt(3))
bfield_tag_dict["pos3-d1"].add(((1, 1,), (1, 1,), (1, 1,),), (0, 1, 1,), 1/sympy.sqrt(3))
bfield_tag_dict["pos3-d1"].add(((1, 1,), (1, 1,), (1, 1,),), (1, 0, 1,), 1/sympy.sqrt(3))

bfield_tag_dict["pos3-d3"] = BfieldCoef()
bfield_tag_dict["pos3-d3"].add(((1, 1,), (1, 1,), (1, 1,),), (1, 1, 1,), 1)

### ------

class Bfield(Op):

    """
    baryon tensor
    #
    self.tag
    self.s1
    self.s2
    self.s3
    self.c1
    self.c2
    self.c3
    #
    tag in bfield_tag_dict
    #
    The tensor is assumed to take the form:
        baryon(s1,s2,s3,c1,c2,c3) = spin_tensor[s1,s2,s3] * q.epsilon_tensor(c1,c2,c3)
        spin_tensor = bfield_tag_dict[self.tag].get_spin_tensor_code()
    """

    def __init__(self, tag:str, s1:str, s2:str, s3:str, c1:str, c2:str, c3:str):
        assert tag in bfield_tag_dict
        Op.__init__(self, "Bfield")
        self.tag = tag
        self.s1 = s1
        self.s2 = s2
        self.s3 = s3
        self.c1 = c1
        self.c2 = c2
        self.c3 = c3

    def __repr__(self) -> str:
        return f"{self.otype}({self.tag!r},{self.s1!r},{self.s2!r},{self.s3!r},{self.c1!r},{self.c2!r},{self.c3!r})"

    def list(self):
        return [ self.otype, self.tag, self.s1, self.s2, self.s3, self.c1, self.c2, self.c3, ]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

### ------

@q.timer
def simplify_bs_elem_list(elem_list:list) -> list:
    """
    Merge `elem_pair` with the same spin settings (combine the `coef`s).
    """
    elem_list = sorted(elem_list, key=repr)
    if len(elem_list) == 0:
        return []
    s_elem_list = []
    value, coef, = elem_list[0]
    for v, c in elem_list[1:]:
        if v == value:
            coef += c
        else:
            coef = ea.simplified_ea(coef)
            if coef != 0:
                s_elem_list.append((value, coef,))
            value, coef, = v, c,
    coef = ea.simplified_ea(coef)
    if coef != 0:
        s_elem_list.append((value, coef,))
    return s_elem_list

@q.timer
def mk_bs_elem_list(tag_v, permute_v, tag_b, permute_b):
    assert isinstance(tag_v, str)
    assert isinstance(tag_b, str)
    assert tag_v in bfield_tag_dict
    assert tag_b in bfield_tag_dict
    assert isinstance(permute_v, tuple)
    assert isinstance(permute_b, tuple)
    assert len(permute_v) == 3
    assert len(permute_b) == 3
    for i in permute_v:
        assert 0 <= i and i < 3
    for i in permute_b:
        assert 0 <= i and i < 3
    v_st = bfield_tag_dict[tag_v].get_spin_tensor_code(permute_v)
    b_st = bfield_tag_dict[tag_b].get_spin_tensor_code(permute_b)
    sst = v_st[:, None, :, None, :, None] * b_st[None, :, None, :, None, :]
    if np.all(sst == 0):
        return
    elem_list = []
    for v_s1 in range(4):
        for v_s2 in range(4):
            for v_s3 in range(4):
                for b_s1 in range(4):
                    for b_s2 in range(4):
                        for b_s3 in range(4):
                            c = sst[v_s1, b_s1, v_s2, b_s2, v_s3, b_s3]
                            if c == 0:
                                continue
                            elem_list.append(((v_s1, b_s1, v_s2, b_s2, v_s3, b_s3,), c,))
    return elem_list

class BS(Op):

    """
    single baryon prop
    #
    self.elem_list
    self.chain_list
    #
    elem_list = [ ((v_s1, b_s1, v_s2, b_s2, v_s3, b_s3,), coef,) ... ]
    chain_list = [ prop_0, prop_1, prop_2, ]
    tag_v or tag_b in `bfield_tag_dict`
    permute_v = (spin_color_index_that_prop_0_contract_with,
                 spin_color_index_that_prop_1_contract_with,
                 spin_color_index_that_prop_2_contract_with,)
    """

    def __init__(self, elem_list:list, chain_list:list[Chain]):
        assert isinstance(elem_list, list)
        for elem in elem_list:
            assert isinstance(elem, tuple)
            assert len(elem) == 2
            assert isinstance(elem[0], tuple)
            ((v_s1, b_s1, v_s2, b_s2, v_s3, b_s3,), coef,) = elem
            assert 0 <= v_s1 and v_s1 < 4
            assert 0 <= b_s1 and b_s1 < 4
            assert 0 <= v_s2 and v_s2 < 4
            assert 0 <= b_s2 and b_s2 < 4
            assert 0 <= v_s3 and v_s3 < 4
            assert 0 <= b_s3 and b_s3 < 4
        assert isinstance(chain_list, list)
        assert len(chain_list) == 3
        for ch in chain_list:
            assert isinstance(ch, Chain)
            assert ch.otype == "Chain"
        Op.__init__(self, "BS")
        self.elem_list = elem_list
        self.chain_list = chain_list

    def add(self, tag:tuple, coef):
        return self.bs_add_tag_coef(tag, coef)

    @q.timer
    def bs_add_tag_coef(self, tag:tuple, coef):
        """
        (tag_v, permute_v, tag_b, permute_b,) = tag
        """
        (tag_v, permute_v, tag_b, permute_b,) = tag
        elem_list = mk_bs_elem_list(tag_v, permute_v, tag_b, permute_b)
        for ss, c in elem_list:
            self.elem_list.append((ss, coef * c))
        self.simplify_elem_list()

    def __imul__(self, factor):
        if factor == 0:
            self.elem_list = []
        elem_list = []
        for elem in self.elem_list:
            (value, coef,) = elem
            elem = (value, coef * factor,)
            elem_list.append(elem)
        self.elem_list = elem_list
        return self

    def __repr__(self) -> str:
        return f"{self.otype}({self.elem_list!r},{self.chain_list!r})"

    def list(self):
        return [ self.otype, self.elem_list, self.chain_list, ]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

    def sort(self):
        sp_chain_list = sorted(list(enumerate(self.chain_list)), key=lambda x: repr(x[1]))
        s_chain_list = []
        i_list = []
        for sp in sp_chain_list:
            i, ch = sp
            i_list.append(i)
            s_chain_list.append(ch)
        def permute_ss(ss):
            pss = []
            for i in i_list:
                pss.append(ss[2 * i])
                pss.append(ss[2 * i + 1])
            return tuple(pss)
        s_elem_list = []
        for elem in self.elem_list:
            (ss, coef,) = elem
            s_elem = (permute_ss(ss), coef,)
            s_elem_list.append(s_elem)
        s_elem_list = sorted(s_elem_list)
        self.elem_list = s_elem_list
        self.chain_list = s_chain_list

    def isospin_symmetric_limit(self) -> None:
        for op in self.chain_list:
            op.isospin_symmetric_limit()

    def get_spin_spin_tensor_code(self) -> np.ndarray|None:
        """
        return sst or None
        sst[v_s1, b_s1, v_s2, b_s2, v_s3, b_s3] = coef
        """
        if len(self.elem_list) == 0:
            return None
        shape = (4, 4, 4, 4, 4, 4,)
        sst = np.zeros(shape, dtype=object)
        for elem in self.elem_list:
            ((v_s1, b_s1, v_s2, b_s2, v_s3, b_s3,), coef,) = elem
            sst[v_s1, b_s1, v_s2, b_s2, v_s3, b_s3] += coef
        return sst

    def get_spin_spin_tensor_elem_list_code(self) -> list:
        """
        return self.elem_list
        self.elem_list = [ ((v_s1, b_s1, v_s2, b_s2, v_s3, b_s3,), coef,), ... ]
        """
        return self.elem_list

    def simplify_elem_list(self):
        self.elem_list = simplify_bs_elem_list(self.elem_list)

### ------

def find_chains_for_bfield_v(bf:Bfield, op_list:list[Op]) -> tuple[list[Chain], list[Op]]|None:
    """
    return chain_list, remaining_op_list
    """
    s_list = [ bf.s1, bf.s2, bf.s3, ]
    c_list = [ bf.c1, bf.c2, bf.c3, ]
    chain_list = [ None for i in range(3) ]
    masks = [ False for op in op_list ]
    for op in op_list:
        if op.otype != "Chain":
            continue
        for i, s in enumerate(s_list):
            if op.s1 == s:
                assert masks[i] == False
                chain_list[i] = op
                masks[i] = True
    assert len(chain_list) == 3
    for i in range(3):
        ch = chain_list[i]
        if ch is None:
            return None
        assert ch is not None
        assert ch.s1 == s_list[i]
        assert ch.c1 == c_list[i]
    remaining_op_list = []
    for mask, op in zip(masks, op_list):
        if mask == False:
            remaining_op_list.append(op)
    return chain_list, remaining_op_list

def find_bfield_b_from_chains(chain_list:list[Chain], op_list:list[Op]) -> tuple[Bfield, tuple[int], list[Op]]|None:
    """
    return (bf_b, permute, remaining_op_list) or None
    """
    assert len(chain_list) == 3
    s_ch_list = [ ch.s2 for ch in chain_list ]
    c_ch_list = [ ch.c2 for ch in chain_list ]
    for idx, op in enumerate(op_list):
        if op.otype != "Bfield":
            continue
        bf = op
        s_list = [ bf.s1, bf.s2, bf.s3, ]
        c_list = [ bf.c1, bf.c2, bf.c3, ]
        if sorted(s_list) != sorted(s_ch_list):
            continue
        assert sorted(c_list) == sorted(c_ch_list)
        permute = []
        for s, c in zip(s_ch_list, c_ch_list):
            i = s_list.index(s)
            assert s == s_list[i]
            assert c == c_list[i]
            permute.append(i)
        permute = tuple(permute)
        remaining_op_list = []
        for i, op in enumerate(op_list):
            if i != idx:
                remaining_op_list.append(op)
        return bf, permute, remaining_op_list
    return None

def mk_baryon_prop(bf_v:Bfield, bf_b:Bfield, chain_list:list[Chain]) -> BS:
    assert len(chain_list) == 3
    assert chain_list[0].s1 != chain_list[1].s1
    assert chain_list[0].s1 != chain_list[2].s1
    assert chain_list[1].s1 != chain_list[2].s1
    assert chain_list[0].s2 != chain_list[1].s2
    assert chain_list[0].s2 != chain_list[2].s2
    assert chain_list[1].s2 != chain_list[2].s2
    assert bf_v.s1 != bf_v.s2
    assert bf_v.s1 != bf_v.s3
    assert bf_v.s2 != bf_v.s3
    assert bf_b.s1 != bf_b.s2
    assert bf_b.s1 != bf_b.s3
    assert bf_b.s2 != bf_b.s3
    ch_list, re_op_list = find_chains_for_bfield_v(bf_v, chain_list)
    assert len(re_op_list) == 0
    assert ch_list == chain_list
    permute_v = (0, 1, 2,)
    bf, permute_b, re_op_list = find_bfield_b_from_chains(chain_list, [ bf_b, ])
    assert bf == bf_b
    assert len(re_op_list) == 0
    tag_v = bf_v.tag
    tag_b = bf_b.tag
    chain_list = [ copy_op_index_auto(ch) for ch in chain_list ]
    bs = BS([], chain_list)
    bs.add((tag_v, permute_v, tag_b, permute_b,), 1)
    return bs

def find_baryon_prop(op_list:list) -> tuple[BS,list[Op]]|None:
    """
    return (baryon_prop, remaining_op_list,) or None
    """
    size = len(op_list)
    for i, op in enumerate(op_list):
        if op.otype != "Bfield":
            continue
        bf_v = op
        remaining_op_list = op_list[:i] + op_list[i + 1:]
        p = find_chains_for_bfield_v(bf_v, remaining_op_list)
        if p is None:
            continue
        chain_list, remaining_op_list = p
        p = find_bfield_b_from_chains(chain_list, remaining_op_list)
        if p is None:
            continue
        bf_b, permute, remaining_op_list = p
        baryon_prop = mk_baryon_prop(bf_v, bf_b, chain_list)
        return baryon_prop, remaining_op_list
    return None

@q.timer
def collect_baryon_props(op_list:list[Op]) -> list[Op]:
    """
    Collect all the `BS`s.
    """
    bs_list = []
    while True:
        p = find_baryon_prop(op_list)
        if p is None:
            break
        bs, op_list = p
        bs_list.append(bs)
    return bs_list + op_list

### ------

class Term:

    """
    self.coef
    self.c_ops
    self.a_ops
    """

    def __init__(self, c_ops, a_ops, coef=1):
        self.coef = coef
        self.c_ops = c_ops
        self.a_ops = a_ops
        for op in c_ops:
            assert op.is_commute()
        for op in a_ops:
            assert not op.is_commute()

    def list(self):
        return [ self.coef, self.c_ops, self.a_ops, ]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

    def check_commute(self) -> bool:
        for op in self.c_ops:
            assert op.is_commute()
        for op in self.a_ops:
            assert not op.is_commute()

    def __imul__(self, factor):
        self.coef *= factor
        return self

    def __add__(self, other):
        return mk_expr(self) + other

    def __add__(self, other):
        return mk_expr(self) + other

    __radd__ = __add__

    def __mul__(self, other):
        return mk_expr(self) * other

    def __rmul__(self, other):
        return mk_expr(other) * self

    def __neg__(self):
        return mk_expr(-1) * mk_expr(self)

    def __pos__(self):
        return self

    def __sub__(self, other):
        return mk_expr(self) + mk_expr(-1) * other

    def __rsub__(self, other):
        return mk_expr(other) + mk_expr(-1) * self

    def show(self, is_multiply=False) -> str:
        return "*".join([ f"({self.coef})", ] + self.c_ops + self.a_ops)

    def __repr__(self) -> str:
        return f"Term({self.c_ops},{self.a_ops},{self.coef})"

    def sort(self) -> None:
        """
        only sort commutable factors
        #
        Important to keep the `BS` factors in front of other factors.
        """
        for op in self.c_ops:
            op.sort()
        self.c_ops.sort(key=repr)

    def simplify_coef(self) -> None:
        self.coef = ea.simplified_coef_ea(self.coef)

    def simplify_ea(self) -> None:
        self.coef = ea.simplified_ea(self.coef)

    @q.timer
    def collect_traces(self) -> None:
        """
        Collect `Chain`s, `Tr`s, `BS`s.
        """
        if len(self.a_ops) == 0:
            self.c_ops = collect_traces(self.c_ops)
            self.c_ops = collect_baryon_props(self.c_ops)

    def isospin_symmetric_limit(self) -> None:
        for op in self.c_ops:
            op.isospin_symmetric_limit()

### ------

class Expr:

    """
    self.description
    self.terms
    """

    def __init__(self, terms, description = None):
        self.description = description
        self.terms = terms

    @q.timer
    def copy(self):
        """
        return a deep copy of this object.
        """
        return copy.deepcopy(self)

    def __imul__(self, factor):
        for term in self.terms:
            term *= factor
        self.description = f"{factor} * {self.show(True)}"
        return self

    def __add__(self, other):
        """
        If other is str, then it is used to set the description of the resulting `expr`.
        Otherwise return self + other.
        """
        if isinstance(other, str):
            return Expr(self.terms, other)
        other = mk_expr(other)
        terms = self.terms + other.terms
        return Expr(terms, f"+{self.show()} + {other.show()}")

    __radd__ = __add__

    def __mul__(self, other):
        other = mk_expr(other)
        terms = []
        for t1 in self.terms:
            for t2 in other.terms:
                coef = t1.coef * t2.coef
                if coef == 0:
                    continue
                t = Term(t1.c_ops + t2.c_ops, t1.a_ops + t2.a_ops, coef)
                terms.append(t)
        return Expr(terms, f"{self.show(True)} * {other.show(True)}")

    def __rmul__(self, other):
        return mk_expr(other) * self

    def __neg__(self):
        return mk_expr(-1) * mk_expr(self)

    def __pos__(self):
        return self

    def __sub__(self, other):
        return mk_expr(self) + mk_expr(-1) * other

    def __rsub__(self, other):
        return mk_expr(other) + mk_expr(-1) * self

    def show(self, is_multiply=False):
        if self.description is not None:
            assert isinstance(self.description, str)
            if len(self.description) >= 1 and self.description[0] == "+":
                if is_multiply:
                    return f"( {self.description[1:]} )"
                else:
                    return f"{self.description[1:]}"
            else:
                return self.description
        else:
            return repr(self)

    def __repr__(self) -> str:
        if self.description is None:
            return f"Expr({self.terms})"
        else:
            return f"Expr({self.terms},{self.description!r})"

    @q.timer
    def sort(self) -> None:
        for term in self.terms:
            term.sort()
        self.terms.sort(key=repr)

    @q.timer
    def simplify_coef(self) -> None:
        for t in self.terms:
            t.simplify_coef()
        self.drop_zeros()

    @q.timer
    def simplify_ea(self) -> None:
        for t in self.terms:
            t.simplify_ea()
        self.drop_zeros()

    @q.timer
    def rescale_bs_term(self) -> None:
        terms = []
        for t in self.terms:
            t = rescale_bs_term(t)
            terms.append(t)
        self.terms = terms

    @q.timer
    def combine_terms(self) -> bool:
        """
        Combine similar terms.
        Return if there are terms be combined.
        """
        expr = combine_terms_expr(self)
        if expr is None:
            return False
        else:
            self.terms = expr.terms
            return True

    @q.timer
    def drop_zeros(self) -> None:
        self.terms = drop_zero_terms(self).terms

    def round(self, ndigit:int=20) -> None:
        """
        interface function
        """
        sexpr = self.copy()
        for term in sexpr.terms:
            coef = term.coef
            term.coef = term.coef.evalf(ndigit)
        sexpr.terms = drop_zero_terms(sexpr).terms
        return sexpr

    @q.timer
    def collect_traces(self) -> None:
        for term in self.terms:
            term.collect_traces()

    @q.timer
    def isospin_symmetric_limit(self) -> None:
        for term in self.terms:
            term.isospin_symmetric_limit()

    @q.timer
    def simplify(self, *, is_isospin_symmetric_limit:bool=True) -> None:
        """
        interface function
        """
        if is_isospin_symmetric_limit:
            self.isospin_symmetric_limit()
        # self.sort()
        # self.combine_terms()
        self.collect_traces()
        self.sort()
        self.combine_terms()
        self.rescale_bs_term()
        self.simplify_ea()
        # self.sort()

### ------

@q.timer
def mk_fac(x) -> Expr:
    """
    interface function
    Stand for "make factor", the result of this function can be used in auto contractor as a factor.
    Make an Expr obj (can be sympy expression).
    `x` can have type `str`, which will be viewed as code segment.
    The code segment can use functions and variables defined in `auto_contractor.auto_fac_funcs`, `position_dict`, `base_position_dict`.
    You can define functions in `position_dict` or `base_position_dict`.
    `position_dict` is argument in function `eval_cexpr`.
    `base_position_dict` is argument in function `cache_compiled_cexpr`.
    """
    return mk_expr(ea.mk_fac(x))

@q.timer
def simplified(expr:Expr, *, is_isospin_symmetric_limit:bool=True) -> Expr:
    """
    interface function
    does not change expr
    """
    sexpr = expr.copy()
    sexpr.simplify(is_isospin_symmetric_limit=is_isospin_symmetric_limit)
    return sexpr

def mk_expr(x) -> Expr:
    if isinstance(x, Op):
        if x.is_commute():
            return Expr([Term([x,], [], 1),], f"{x}")
        else:
            return Expr([Term([], [x,], 1),], f"{x}")
    elif isinstance(x, Term):
        return Expr([x,], f"x.show()")
    elif isinstance(x, Expr):
        return x
    elif isinstance(x, (int, float, complex, sympy.Basic, ea.Expr)):
        return Expr([Term([], [], x),], f"({x})")
    else:
        raise Exception(f"{x}")

def get_op_signature(op:Op, bs_count:int) -> str:
    if bs_count == 0:
        if op.otype == "BS":
            return f"{op.otype}(elem_list,{op.chain_list!r})"
    return f"{op!r}"

def get_term_signature(term:Term) -> str:
    """
    Terms with the same signature can be combined together.
    """
    c_ops_str_list = []
    bs_count = 0
    for op in term.c_ops:
        c_ops_str_list.append(get_op_signature(op, bs_count))
        if op.otype == "BS":
            bs_count += 1
    c_ops_str = ",".join(c_ops_str_list)
    return (f"[{c_ops_str}],{term.a_ops!r}", bs_count,)

def get_bs_list_from_op_list(op_list:list[Op]) -> tuple[list[BS], list[Op]]:
    bs_list = []
    remaining_op_list = []
    for op in op_list:
        if op.otype == "BS":
            bs_list.append(op)
        else:
            remaining_op_list.append(op)
    return bs_list, remaining_op_list

@q.timer
def combine_two_terms(t1:Term, t2:Term, t1_sig:str, t2_sig:str) -> Term|None:
    """
    Combine two terms together.
    If not possible, return None.
    """
    if t1_sig == t2_sig:
        (sig_str, bs_count,) = t1_sig
        assert t1.a_ops == t2.a_ops
        if bs_count == 0:
            assert t1.c_ops == t2.c_ops
            coef = t1.coef + t2.coef
            if ea.is_zero(coef):
                return Term([], [], 0)
            else:
                return Term(t1.c_ops, t1.a_ops, coef)
        elif bs_count == 1:
            coef1 = t1.coef
            coef2 = t2.coef
            [ bs1, ], re_c_ops1, = get_bs_list_from_op_list(t1.c_ops)
            [ bs2, ], re_c_ops2, = get_bs_list_from_op_list(t2.c_ops)
            assert re_c_ops1 == re_c_ops2
            assert bs1.chain_list == bs2.chain_list
            bs1 = copy.copy(bs1)
            bs2 = copy.copy(bs2)
            bs1 *= coef1
            bs2 *= coef2
            bs = BS(bs1.elem_list + bs2.elem_list, bs1.chain_list)
            bs.simplify_elem_list()
            c_ops = [ bs, ] + re_c_ops1
            return Term(c_ops, t1.a_ops, 1)
        else:
            return None
    else:
        return None

@q.timer
def combine_terms_expr(expr:Expr) -> Expr|None:
    """
    Combine terms with the same signatures.
    Return `None` if not terms are combined.
    """
    if not expr.terms:
        return None
    signatures = [ get_term_signature(t) for t in expr.terms ]
    if len(signatures) == len(set(signatures)):
        return None
    zero_term = Term([], [], 0)
    zero_term_sig = get_term_signature(zero_term)
    s_pairs = sorted(list(zip(expr.terms, signatures)), key=lambda x: x[1])
    s_terms, s_signatures, = list(zip(*s_pairs))
    terms = []
    term = s_terms[0]
    term_sig = s_signatures[0]
    for t, t_sig in zip(s_terms[1:], s_signatures[1:]):
        if ea.is_zero(term.coef):
            term = t
            term_sig = t_sig
        else:
            ct = combine_two_terms(t, term, t_sig, term_sig)
            if ct is None:
                terms.append(term)
                term = t
                term_sig = t_sig
            elif ea.is_zero(ct.coef):
                term = zero_term
                term_sig = zero_term_sig
            else:
                term = ct
                assert term_sig == get_term_signature(term)
    if not ea.is_zero(term.coef):
        terms.append(term)
    return Expr(terms, expr.description)

def rescale_bs_term(term:Term) -> Term:
    bs_list, re_op_list, = get_bs_list_from_op_list(term.c_ops)
    if len(bs_list) == 0:
        return term
    scaled_bs_list = []
    coef_prod = 1
    for bs in bs_list:
        sst_el = bs.get_spin_spin_tensor_elem_list_code()
        over_all_coef = 0
        for _, coef, in sst_el:
            if isinstance(coef, ea.Expr):
                coef = ea.simplified_ea(coef)
            if ea.is_zero(coef):
                continue
            if isinstance(coef, ea.Expr):
                if coef.is_zero():
                    continue
                over_all_coef = coef.terms[0].coef
                break
            else:
                if coef == 0:
                    continue
                over_all_coef = coef
                break
        if over_all_coef == 0:
            return Term([], [], 0)
        scaled_bs = copy.copy(bs)
        scaled_bs *= 1 / mk_sym(over_all_coef)
        scaled_bs_list.append(scaled_bs)
        coef_prod *= over_all_coef
    scaled_term = Term(scaled_bs_list + re_op_list, term.a_ops, coef_prod * term.coef)
    return scaled_term

@q.timer
def drop_zero_terms(expr:Expr) -> Expr:
    terms = []
    for t in expr.terms:
        if ea.is_zero(t.coef):
            continue
        terms.append(t)
    return Expr(terms, expr.description)

def op_derivative_exp(op:Op):
    if op.otype == "Qv":
        return Term([], [SHv(op.f, op.p, op.s, op.c),], -1)
    elif op.otype == "Qb":
        return Term([], [HbS(op.f, op.p, op.s, op.c),], 1)
    else:
        return None

def op_derivative_op(op:Op, op1:Op):
    if op.otype == "Qv" and op1.otype == "HbS" and op.f == op1.f:
        return Term([S(op.f, op.p, op1.p, op.s, op1.s, op.c, op1.c),], [], 1)
    elif op.otype == "Qb" and op1.otype == "SHv" and op.f == op1.f:
        return Term([S(op.f, op1.p, op.p, op1.s, op.s, op1.c, op.c),], [], 1)
    else:
        return None

def flip_sign(i:int) -> int:
    if i % 2 == 0:
        return 1
    else:
        return -1

def op_derivative_term(op:Op, term:Term) -> Expr:
    coef = term.coef
    c_ops = term.c_ops
    a_ops = term.a_ops
    terms = []
    for i in range(len(a_ops)):
        op1 = a_ops[i]
        dop1 = op_derivative_op(op, op1)
        if dop1 is not None:
            sign = flip_sign(i)
            terms.append(Term(dop1.c_ops + c_ops, a_ops[:i] + dop1.a_ops + a_ops[i+1:], sign * dop1.coef * coef))
    de = op_derivative_exp(op)
    if de is not None:
        sign = flip_sign(len(a_ops))
        terms.append(Term(de.c_ops + c_ops, a_ops + de.a_ops, sign * de.coef * coef))
    return Expr(terms)

def op_push_term(op:Op, term:Term) -> Expr:
    if op.otype == "Qv" or op.otype == "Qb":
        return op_derivative_term(op, term)
    else:
        coef = term.coef
        c_ops = term.c_ops
        a_ops = term.a_ops
        if op.is_commute():
            return Expr([Term([op,] + c_ops, a_ops, coef),])
        else:
            return Expr([Term(c_ops, [op,] + a_ops, coef),])

def op_push_expr(op:Op, expr:Expr) -> Expr:
    terms = []
    for term in expr.terms:
        terms += op_push_term(op, term).terms
    return Expr(terms)

def is_hop(op:Op) -> bool:
    if op.otype == "SHv" or op.otype == "HbS":
        return True
    if op.otype == "Hv" or op.otype == "Hb":
        return True
    return False

def has_hops(term:Term, count_limit:int = 0) -> bool:
    c = 0
    for op in term.a_ops:
        if is_hop(op):
            c += 1
            if c > count_limit:
                return True
    return False

def remove_hops(expr:Expr, count_limit:int = 0) -> Expr:
    terms = []
    for term in expr.terms:
        if not has_hops(term, count_limit):
            terms.append(term)
    return Expr(terms)

def contract_term(term:Term) -> Expr:
    coef = term.coef
    c_ops = term.c_ops
    a_ops = term.a_ops
    n_a_ops = len(a_ops)
    expr = Expr([Term(c_ops, [], coef),])
    for idx, op in enumerate(reversed(a_ops)):
        expr = op_push_expr(op, expr)
        expr = remove_hops(expr, n_a_ops - idx - 1)
    return expr

@q.timer
def contract_expr(expr:Expr) -> Expr:
    """
    interface function
    does not change expr
    """
    all_terms = []
    for term in expr.terms:
        all_terms += contract_term(term).terms
    return Expr(all_terms, f"< {expr.show()} >")

### ------

def S_l(p1, p2):
    return mk_expr(S('l', p1, p2)) + f"S_l({p1},{p2})"

def S_s(p1, p2):
    return mk_expr(S('s', p1, p2)) + f"S_s({p1},{p2})"

def S_c(p1, p2):
    return mk_expr(S('c', p1, p2)) + f"S_c({p1},{p2})"

def tr(expr):
    if isinstance(expr, Term):
        term = expr
        assert term.a_ops == []
        return Term([ Tr(term.c_ops), ], [], term.coef)
    elif isinstance(expr, Expr):
        return sum(map(tr, expr.terms)) + f"tr( {expr.show()} )"
    else:
        assert False

gamma_x = mk_expr(G(0)) + f"gamma_x"

gamma_y = mk_expr(G(1)) + f"gamma_y"

gamma_z = mk_expr(G(2)) + f"gamma_z"

gamma_t = mk_expr(G(3)) + f"gamma_t"

gamma_5 = mk_expr(G(5)) + f"gamma_5"

def gamma(tag):
    return mk_expr(G(tag)) + f"gamma({tag})"

def gamma_va(tag):
    assert isinstance(tag, int)
    if tag in [ 0, 1, 2, 3, ]:
        return mk_expr(G(tag)) + f"gamma({tag})"
    elif tag in [ 4, 5, 6, 7, ]:
        return mk_expr(G(tag - 4)) * mk_expr(G(5)) + f"gamma({tag-4})*gamma_5"
    else:
        assert False

### ------

def mk_test_expr_wick_01():
    expr = (1
            * Qb("d", "x1", "s1", "c1")
            * G(5, "s1", "s2")
            * Qv("u", "x1", "s2", "c1")
            * Qb("u", "x2", "s3", "c2")
            * G(5, "s3", "s4")
            * Qv("d", "x2", "s4", "c2"))
    return expr

def mk_test_expr_wick_02():
    f1 = "u"
    f2 = "u"
    f3 = "d"
    #
    p1 = "x1"
    s1 = "s1"
    c1 = "c1"
    #
    p2 = "x2"
    s2 = "s2"
    c2 = "c2"
    #
    p3 = "x3"
    s3 = "s3"
    c3 = "c3"
    #
    p1p = "x1p"
    s1p = "s1p"
    c1p = "c1p"
    #
    p2p = "x2p"
    s2p = "s2p"
    c2p = "c2p"
    #
    p3p = "x3p"
    s3p = "s3p"
    c3p = "c3p"
    #
    q1v = Qv(f1, p1p, s1p, c1p)
    q2v = Qv(f2, p2p, s2p, c2p)
    q3v = Qv(f3, p3p, s3p, c3p)
    q1b = Qb(f1, p1, s1, c1)
    q2b = Qb(f2, p2, s2, c2)
    q3b = Qb(f3, p3, s3, c3)
    #
    bf_b = Bfield("std-u", s1, s2, s3, c1, c2, c3)
    bf_v = Bfield("std-u", s1p, s2p, s3p, c1p, c2p, c3p)
    #
    expr = bf_b * bf_v * q1v * q2v * q3v * q1b * q2b * q3b + f"expr"
    return expr

def mk_test_expr_wick_03():
    f1 = "d"
    f2 = "u"
    f3 = "u"
    #
    p1 = "x1"
    s1 = "s1"
    c1 = "c1"
    #
    p2 = "x2"
    s2 = "s2"
    c2 = "c2"
    #
    p3 = "x2"
    s3 = "s3"
    c3 = "c3"
    #
    p1p = "x1p"
    s1p = "s1p"
    c1p = "c1p"
    #
    p2p = "x2p"
    s2p = "s2p"
    c2p = "c2p"
    #
    p3p = "x2p"
    s3p = "s3p"
    c3p = "c3p"
    #
    q1v = Qv(f1, p1p, s1p, c1p)
    q2v = Qv(f2, p2p, s2p, c2p)
    q3v = Qv(f3, p3p, s3p, c3p)
    q1b = Qb(f1, p1, s1, c1)
    q2b = Qb(f2, p2, s2, c2)
    q3b = Qb(f3, p3, s3, c3)
    #
    bf_b = Bfield("std-u", s1, s2, s3, c1, c2, c3)
    bf_v = Bfield("std-u", s1p, s2p, s3p, c1p, c2p, c3p)
    #
    expr = bf_b * bf_v * q1v * q2v * q3v * q1b * q2b * q3b + f"expr"
    return expr

def mk_test_expr_wick_04():
    f1 = "d"
    f2 = "u"
    f3 = "u"
    #
    p1 = "x1"
    s1 = "s1"
    c1 = "c1"
    #
    p2 = "x2"
    s2 = "s2"
    c2 = "c2"
    #
    p3 = "x2"
    s3 = "s3"
    c3 = "c3"
    #
    p1p = "x1p"
    s1p = "s1p"
    c1p = "c1p"
    #
    p2p = "x2p"
    s2p = "s2p"
    c2p = "c2p"
    #
    p3p = "x2p"
    s3p = "s3p"
    c3p = "c3p"
    #
    q1v = Qv(f1, p1p, s1p, c1p)
    q2v = Qv(f2, p2p, s2p, c2p)
    q3v = Qv(f3, p3p, s3p, c3p)
    q1b = Qb(f1, p1, s1, c1)
    q2b = Qb(f2, p2, s2, c2)
    q3b = Qb(f3, p3, s3, c3)
    #
    bf_b = Bfield("std-u", s1, s2, s3, c1, c2, c3)
    bf_v = Bfield("std-u", s1p, s2p, s3p, c1p, c2p, c3p)
    #
    ubar_u = Qb(f2, "xx", "s01", "c01") * Qv(f2, "xx", "s01", "c01")
    #
    expr = ubar_u * bf_b * bf_v * q1v * q2v * q3v * q1b * q2b * q3b + f"expr"
    return expr

def mk_test_expr_wick_05():
    f1 = "u"
    f2 = "u"
    f3 = "d"
    #
    p1 = "x1"
    s1 = "s1"
    c1 = "c1"
    #
    p2 = "x1"
    s2 = "s2"
    c2 = "c2"
    #
    p3 = "x1"
    s3 = "s3"
    c3 = "c3"
    #
    p1p = "x1p"
    s1p = "s1p"
    c1p = "c1p"
    #
    p2p = "x1p"
    s2p = "s2p"
    c2p = "c2p"
    #
    p3p = "x1p"
    s3p = "s3p"
    c3p = "c3p"
    #
    q1v = Qv(f1, p1p, s1p, c1p)
    q2v = Qv(f2, p2p, s2p, c2p)
    q3v = Qv(f3, p3p, s3p, c3p)
    q1b = Qb(f1, p1, s1, c1)
    q2b = Qb(f2, p2, s2, c2)
    q3b = Qb(f3, p3, s3, c3)
    #
    bf_b = Bfield("std-u", s1, s2, s3, c1, c2, c3)
    # bf_v = Bfield("std-u", s1p, s2p, s3p, c1p, c2p, c3p)
    bf_v = Bfield("std-u", s3p, s2p, s1p, c3p, c2p, c1p)
    #
    expr = bf_b * bf_v * q1v * q2v * q3v * q1b * q2b * q3b + f"expr"
    return expr

def mk_test_expr_wick_06():
    f1 = "s"
    f2 = "s"
    f3 = "s"
    #
    p1 = "x1"
    s1 = "s1"
    c1 = "c1"
    #
    p2 = "x1"
    s2 = "s2"
    c2 = "c2"
    #
    p3 = "x1"
    s3 = "s3"
    c3 = "c3"
    #
    p1p = "x1p"
    s1p = "s1p"
    c1p = "c1p"
    #
    p2p = "x1p"
    s2p = "s2p"
    c2p = "c2p"
    #
    p3p = "x1p"
    s3p = "s3p"
    c3p = "c3p"
    #
    q1v = Qv(f1, p1p, s1p, c1p)
    q2v = Qv(f2, p2p, s2p, c2p)
    q3v = Qv(f3, p3p, s3p, c3p)
    q1b = Qb(f1, p1, s1, c1)
    q2b = Qb(f2, p2, s2, c2)
    q3b = Qb(f3, p3, s3, c3)
    #
    expr_list = []
    #
    bf_b = Bfield("std3-u3", s1, s2, s3, c1, c2, c3)
    bf_v = Bfield("std3-u3", s1p, s2p, s3p, c1p, c2p, c3p)
    #
    expr = bf_b * bf_v * q1v * q2v * q3v * q1b * q2b * q3b + f"expr-std3-u3"
    expr_list.append(expr)
    #
    bf_b = Bfield("std3-u1", s1, s2, s3, c1, c2, c3)
    bf_v = Bfield("std3-u1", s1p, s2p, s3p, c1p, c2p, c3p)
    #
    expr = bf_b * bf_v * q1v * q2v * q3v * q1b * q2b * q3b + f"expr-std3-u1"
    expr_list.append(expr)
    #
    bf_b = Bfield("std3-d1", s1, s2, s3, c1, c2, c3)
    bf_v = Bfield("std3-d1", s1p, s2p, s3p, c1p, c2p, c3p)
    #
    expr = bf_b * bf_v * q1v * q2v * q3v * q1b * q2b * q3b + f"expr-std3-d1"
    expr_list.append(expr)
    #
    bf_b = Bfield("std3-d3", s1, s2, s3, c1, c2, c3)
    bf_v = Bfield("std3-d3", s1p, s2p, s3p, c1p, c2p, c3p)
    #
    expr = bf_b * bf_v * q1v * q2v * q3v * q1b * q2b * q3b + f"expr-std3-d3"
    expr_list.append(expr)
    #
    bf_v = Bfield("std3-u3", s1p, s2p, s3p, c1p, c2p, c3p)
    bf_b = Bfield("std3-u3", s1, s2, s3, c1, c2, c3)
    #
    expr = bf_b * bf_v * q1v * q2v * q3v * q1b * q2b * q3b + f"expr-std3-std3-u3"
    expr_list.append(expr)
    #
    bf_v = Bfield("pos3-u3", s1p, s2p, s3p, c1p, c2p, c3p)
    bf_b = Bfield("pos3-u3", s1, s2, s3, c1, c2, c3)
    #
    expr = bf_b * bf_v * q1v * q2v * q3v * q1b * q2b * q3b + f"expr-pos3-pos3-u3"
    expr_list.append(expr)
    #
    bf_v = Bfield("pos3-u3", s1p, s2p, s3p, c1p, c2p, c3p)
    bf_b = Bfield("std3-u3", s1, s2, s3, c1, c2, c3)
    #
    expr = bf_b * bf_v * q1v * q2v * q3v * q1b * q2b * q3b + f"expr-pos3-std3-u3"
    expr_list.append(expr)
    #
    bf_v = Bfield("std3-u3", s1p, s2p, s3p, c1p, c2p, c3p)
    bf_b = Bfield("pos3-u3", s1, s2, s3, c1, c2, c3)
    #
    from operators import mk_j_mu
    mu = 2
    fac = mk_fac(f"rel_mod_sym(x1[1][{mu}] - x1p[1][{mu}], size[{mu}])")
    #
    expr = fac * mk_j_mu("xx_1", 3) * bf_b * bf_v * q1v * q2v * q3v * q1b * q2b * q3b + f"expr-std3-pos3-u3"
    expr_list.append(expr)
    #
    return expr_list

def mk_test_expr_wick_07():
    f1 = "s"
    f2 = "s"
    f3 = "s"
    #
    p1 = "x1"
    s1 = "s1"
    c1 = "c1"
    #
    p2 = "x1"
    s2 = "s2"
    c2 = "c2"
    #
    p3 = "x1"
    s3 = "s3"
    c3 = "c3"
    #
    p1p = "x1p"
    s1p = "s1p"
    c1p = "c1p"
    #
    p2p = "x1p"
    s2p = "s2p"
    c2p = "c2p"
    #
    p3p = "x1p"
    s3p = "s3p"
    c3p = "c3p"
    #
    q1v = Qv(f1, p1p, s1p, c1p)
    q2v = Qv(f2, p2p, s2p, c2p)
    q3v = Qv(f3, p3p, s3p, c3p)
    q1b = Qb(f1, p1, s1, c1)
    q2b = Qb(f2, p2, s2, c2)
    q3b = Qb(f3, p3, s3, c3)
    #
    expr_list = []
    #
    bf_v = Bfield("std3-u3", s1p, s2p, s3p, c1p, c2p, c3p)
    bf_b = Bfield("std3-u3", s1, s2, s3, c1, c2, c3)
    #
    from operators import mk_j_mu
    expr = mk_j_mu("xx_1", 3) * mk_j_mu("xx_2", 3) * bf_b * bf_v * q1v * q2v * q3v * q1b * q2b * q3b + f"expr-std3-std3-u3"
    expr_list.append(expr)
    #
    return expr_list

if __name__ == "__main__":
    expr = mk_test_expr_wick_01()
    print(expr)
    c_expr = contract_expr(expr)
    c_expr.simplify(is_isospin_symmetric_limit = False)
    print(c_expr)
    c_expr_check = Expr([Term([Tr([G('5'), S('d','x2','x1'), G('5'), S('u','x1','x2')],'sc')],[],(-1+0j))]) - c_expr
    c_expr_check.simplify(is_isospin_symmetric_limit = False)
    print(c_expr_check)
    c_expr.simplify(is_isospin_symmetric_limit = True)
    print(c_expr)
    print(Qb("u", "x2", "s23", "c23") * Qv("u", "x1", "s11", "c11") - Qb("d", "x2", "s23", "c23") * Qv("d", "x1", "s11", "c11"))
    print(bfield_tag_dict['std-u'].get_spin_tensor_code())
    print(bfield_tag_dict['std-u'])
    expr = mk_test_expr_wick_02()
    print(expr)
    c_expr = contract_expr(expr)
    print(c_expr)
    c_expr.simplify()
    print(c_expr)
    q.timer_display()



from qlat_utils import *
from .c import *
from . import c

class QMAction:

    def __init__(self, alpha, beta, V_FV_min, FV_offset, TV_offset, barrier_strength, L, M, epsilon, t_FV_out, t_FV_mid, dt, measure_offset_L, measure_offset_M):
        self.cdata = c.mk_qm_action(alpha, beta, V_FV_min, FV_offset, TV_offset, barrier_strength, L, M, epsilon, t_FV_out, t_FV_mid, dt, measure_offset_L, measure_offset_M)

    def __del__(self):
        assert isinstance(self.cdata, int)
        c.free_qm_action(self)

    def __imatmul__(self, v1):
        assert isinstance(v1, QMAction)
        c.set_qm_action(self, v1)
        return self

    def alpha(self):
        return c.get_alpha_qm_action(self)

    def beta(self):
        return c.get_beta_qm_action(self)

    def barrier_strength(self):
        return c.get_barrier_strength_qm_action(self)

    def M(self):
        return c.get_M_qm_action(self)

    def L(self):
        return c.get_L_qm_action(self)

    def t_FV_out(self):
        return c.get_t_FV_out_qm_action(self)

    def t_FV_mid(self):
        return c.get_t_FV_mid_qm_action(self)
    
    def t_FV(self):
        return 2*self.t_FV_out() + self.t_FV_mid()

    def dt(self):
        return c.get_dt_qm_action(self)

    def V(self, x, t):
        return c.V_qm_action(self, x[0], x[1], t)

    def dV(self, x, t):
        return c.dV_qm_action(self, x[0], x[1], t)

    def action_node(self, f):
        assert isinstance(f, FieldBase)
        return c.action_node_qm_action(self, f)

    def hmc_m_hamilton_node(self, m):
        assert isinstance(m, FieldBase)
        return c.hmc_m_hamilton_node_qm_action(self, m)

    def sum_sq(self, f):
        assert isinstance(f, FieldBase)
        return c.sum_sq_qm_action(self, f)

    def hmc_set_force(self, force, f):
        assert isinstance(force, FieldBase)
        assert isinstance(f, FieldBase)
        return c.hmc_set_force_qm_action(self, force, f)

    def hmc_field_evolve(self, f, m, step_size):
        assert isinstance(f, FieldBase)
        assert isinstance(m, FieldBase)
        c.hmc_field_evolve_qm_action(self, f, m, step_size)

    def hmc_set_rand_momentum(self, m, rs):
        assert isinstance(m, FieldBase)
        assert isinstance(rs, RngState)
        return c.hmc_set_rand_momentum_qm_action(self, m, rs)



# cython: binding=True, embedsignature=True, c_string_type=unicode, c_string_encoding=utf8

from qlat_utils.all cimport *
from . cimport everything as cc
from .geometry cimport Geometry
from .field_types cimport (
    FieldRealD,
    FieldColorMatrix,
    FieldComplexD,
)
from .propagator cimport (
    Prop,
    SelProp,
    PselProp,
    FermionField4d,
    SpinProp,
)
from .field_selection cimport (
    FieldSelection,
    PointsSelection,
)

from .field_utils import (
    field_expanded,
    refresh_expanded_1,
)

from cpython cimport Py_buffer
from cpython.buffer cimport PyBUF_FORMAT

import cqlat as c
import qlat_utils as q
import numpy as np

cdef class GaugeField(FieldColorMatrix):

    def __init__(self, Geometry geo=None, multiplicity=4):
        assert multiplicity == 4
        super().__init__(geo, 4)

    cdef cc.Handle[cc.GaugeField] xxx(self):
        assert self.xx.multiplicity == 4 or self.xx.multiplicity == 0
        return cc.Handle[cc.GaugeField](<cc.GaugeField&>self.xx)

    @q.timer
    def save(self, path):
        """
        Save with the standard NERSC format
        """
        return c.save_gauge_field(self, path)

    @q.timer
    def load(self, path):
        """
        Load with the standard NERSC format
        """
        return c.load_gauge_field(self, path)

    def set_rand(self, RngState rng, cc.RealD sigma=0.5, cc.Int n_step=1):
        set_g_rand_color_matrix_field(self, rng, sigma, n_step)

    def unitarize(self):
        cc.unitarize(self.xx)

    def plaq(self):
        return gf_avg_plaq(self)

    def link_trace(self):
        return gf_avg_link_trace(self)

    def twist_boundary_at_boundary(self, cc.RealD lmom=-0.5, cc.Int mu=3):
        """
        modify in place
        """
        gf_twist_boundary_at_boundary(self, lmom, mu)

    def show_info(self):
        gf_show_info(self)

###

cdef class GaugeTransform(FieldColorMatrix):

    def __init__(self, Geometry geo=None, multiplicity=1):
        assert multiplicity == 1
        super().__init__(geo, 1)

    cdef cc.Handle[cc.GaugeTransform] xxx(self):
        assert self.xx.multiplicity == 1 or self.xx.multiplicity == 0
        return cc.Handle[cc.GaugeTransform](<cc.GaugeTransform&>self.xx)

    @q.timer
    def save(self, path):
        """
        Save as RealD precision with the generic Field format ``save_double``
        """
        return self.save_double(path)

    @q.timer
    def load(self, path):
        """
        Load as RealD precision with the generic Field format ``load_double``
        """
        return self.load_double(path)

    @q.timer
    def save_cps(self, path):
        """
        Save with the format used in CPS
        """
        return c.save_gauge_transform_cps(self, path)

    @q.timer
    def load_cps(self, path):
        """
        Load with the format used in CPS
        """
        return c.load_gauge_transform_cps(self, path)

    def set_rand(self, RngState rng, cc.RealD sigma=0.5, cc.Int n_step=1):
        set_g_rand_color_matrix_field(self, rng, sigma, n_step)

    def unitarize(self):
        cc.unitarize(self.xx)

    def __mul__(self, other):
        """
        other can be GaugeTransform, GaugeField, Prop, SelProp, PselProp, list
        """
        cdef GaugeTransform gt, gt1
        cdef GaugeField gf, gf1
        cdef Prop prop, prop1
        cdef SelProp s_prop, s_prop1
        cdef PselProp ps_prop, ps_prop1
        cdef FermionField4d ff, ff1
        cdef FieldSelection fsel
        cdef PointsSelection psel
        if isinstance(other, GaugeTransform):
            gt1 = other
            gt = GaugeTransform()
            cc.gt_apply_gauge_transformation(gt.xxx().val(), gt1.xxx().val(), self.xxx().val())
            return gt
        elif isinstance(other, GaugeField):
            gf1 = other
            gf = GaugeField()
            cc.gf_apply_gauge_transformation(gf.xxx().val(), gf1.xxx().val(), self.xxx().val(), False)
            return gf
        elif isinstance(other, Prop):
            prop1 = other
            prop = Prop()
            cc.prop_apply_gauge_transformation(prop.xxx().val(), prop1.xxx().val(), self.xxx().val())
            return prop
        elif isinstance(other, SelProp):
            s_prop1 = other
            fsel = other.fsel
            s_prop = SelProp(fsel)
            cc.prop_apply_gauge_transformation(s_prop.xx, s_prop1.xx, self.xxx().val(), fsel.xx)
            return s_prop
        elif isinstance(other, PselProp):
            ps_prop1 = other
            psel = other.psel
            ps_prop = PselProp(psel)
            cc.prop_apply_gauge_transformation(ps_prop.xx, ps_prop1.xx, self.xxx().val(), psel.xx)
            return ps_prop
        elif isinstance(other, FermionField4d):
            ff1 = other
            ff = FermionField4d()
            cc.ff_apply_gauge_transformation(ff.xxx().val(), ff1.xxx().val(), self.xxx().val())
            return ff
        elif isinstance(other, list):
            return [ self * p for p in other ]
        else:
            raise Exception("GaugeTransform.__mul__")

    def inv(self):
        cdef GaugeTransform gt_inv = GaugeTransform()
        cc.gt_invert(gt_inv.xxx().val(), self.xxx().val())
        return gt_inv

###

@q.timer_verbose
def gf_show_info(GaugeField gf):
    assert gf is not None
    q.displayln_info(f"gf_show_info: plaq = {gf.plaq():.16F} ; link_trace = {gf.link_trace():.16F}.")

def gf_avg_plaq(GaugeField gf):
    assert gf is not None
    return cc.gf_avg_plaq(gf.xxx().val())

def gf_avg_spatial_plaq(GaugeField gf):
    assert gf is not None
    return cc.gf_avg_spatial_plaq(gf.xxx().val())

def gf_avg_link_trace(GaugeField gf):
    assert gf is not None
    return cc.gf_avg_link_trace(gf.xxx().val())

def gf_plaq_field(GaugeField gf):
    assert gf is not None
    cdef FieldRealD f_plaq = FieldRealD()
    cc.gf_plaq_field(f_plaq.xx, gf.xxx().val())
    return f_plaq

@q.timer
def set_tr_less_anti_herm_matrix(FieldColorMatrix fc):
    cc.make_tr_less_anti_herm_matrix(fc.xx)

def gf_wilson_line_no_comm(wlf, m, gf_ext, path, path_n=None):
    """
    wlf = FieldColorMatrix(geo)
    will only modify the m'th component
    e.g. path = [ mu, mu, nu, -mu-1, -mu-1, ]
    e.g. path = [ mu, nu, -mu-1, ], path_n = [ 2, 1, 2, ]
    """
    if path_n is None:
        c.gf_wilson_line_no_comm(wlf, m, gf_ext, path)
    else:
        c.gf_wilson_line_no_comm(wlf, m, gf_ext, path, path_n)

def gf_wilson_lines_no_comm(gf_ext, path_list):
    """
    path_list = [ path_spec, ... ]
    e.g. path_spec = [ mu, mu, nu, -mu-1, -mu-1, ]
    e.g. path_spec = ([ mu, nu, -mu-1, ], [ 2, 1, 2, ],)
    return wlf
    """
    multiplicity = len(path_list)
    geo = q.geo_resize(gf_ext.geo)
    wlf = FieldColorMatrix(geo, multiplicity)
    for m, p in enumerate(path_list):
        if isinstance(p, tuple) and len(p) == 2:
            path, path_n = p
            gf_wilson_line_no_comm(wlf, m, gf_ext, path, path_n)
        else:
            path = p
            gf_wilson_line_no_comm(wlf, m, gf_ext, path)
    return wlf

def gf_avg_wilson_loop_normalized_tr(gf, l, t):
    assert isinstance(gf, GaugeField)
    assert isinstance(l, int)
    assert isinstance(t, int)
    return c.gf_avg_wilson_loop_normalized_tr(gf, l, t)

def set_g_rand_color_matrix_field(fc, rng, sigma, n_steps=1):
    assert isinstance(fc, FieldColorMatrix)
    assert isinstance(rng, RngState)
    return c.set_g_rand_color_matrix_field(fc, rng, sigma, n_steps)

def gf_twist_boundary_at_boundary(GaugeField gf, cc.RealD lmom=-0.5, int mu=3):
    """
    modify gf in place
    """
    c.gf_twist_boundary_at_boundary(gf, lmom, mu)

def mk_left_expanded_field(gf):
    """
    Return left expanded field.
    Similar to ``set_left_expanded_gauge_field`` in C++
    """
    gf1 = field_expanded(gf, 1, 0)
    refresh_expanded_1(gf1)
    return gf1

def gf_reduce_half(GaugeField gf):
    """
    return hgf with half of the size in all directions
    Do not modify `gf`.
    Can perform some shift before call this function, e.g.
    `gf_reduce_half(gf.shift())`
    """
    cdef GaugeField hgf = GaugeField()
    cc.gf_reduce_half(hgf.xxx().val(), gf.xxx().val())
    return hgf

###

@q.timer
def invert_dwf_qed(
        FieldComplexD f_in4d, FieldComplexD gf1,
        cc.RealD mass, cc.RealD m5, cc.Int ls,
        *,
        t_wick_phase_factor_arr=None,
        cc.Bool is_dagger=False,
        cc.RealD stop_rsd=1e-8, cc.Long max_num_iter=50000,
        ):
    """
    properly project to 4d fermion field
    if is_dagger is false (default), then M out = in
    if is_dagger is true, then M^dag out = in
    #
    gf1 = q.mk_left_expanded_field(gf)
    """
    cdef FieldComplexD f_out4d = FieldComplexD()
    cdef cc.vector[cc.ComplexD] t_wick_phase_factor_vec = cc.vector[cc.ComplexD]()
    cdef cc.Int t_size
    cdef cc.Int i
    if t_wick_phase_factor_arr is not None:
        t_size = len(t_wick_phase_factor_arr)
        t_wick_phase_factor_vec.resize(t_size)
        for i in range(t_size):
            t_wick_phase_factor_vec[i] = cc.ccpy_d(t_wick_phase_factor_arr[i])
    cc.invert_dwf_qed(
        f_out4d.xx, f_in4d.xx, gf1.xx,
        mass, m5, ls, t_wick_phase_factor_vec, is_dagger, stop_rsd, max_num_iter)
    return f_out4d

@q.timer
def cg_with_m_dwf_qed(
        FieldComplexD f_in5d, FieldComplexD gf1,
        cc.RealD mass, cc.RealD m5, cc.Int ls,
        *,
        t_wick_phase_factor_arr=None,
        cc.Bool is_dagger=False,
        cc.RealD stop_rsd=1e-8, cc.Long max_num_iter=50000,
        ):
    """
    if is_dagger is false, then M^dag M out = in
    if is_dagger is true, then M M^dag out = in
    #
    gf1 = q.mk_left_expanded_field(gf)
    """
    cdef FieldComplexD f_out5d = FieldComplexD()
    cdef cc.vector[cc.ComplexD] t_wick_phase_factor_vec = cc.vector[cc.ComplexD]()
    cdef cc.Int t_size
    cdef cc.Int i
    if t_wick_phase_factor_arr is not None:
        t_size = len(t_wick_phase_factor_arr)
        t_wick_phase_factor_vec.resize(t_size)
        for i in range(t_size):
            t_wick_phase_factor_vec[i] = cc.ccpy_d(t_wick_phase_factor_arr[i])
    cc.cg_with_m_dwf_qed(
        f_out5d.xx, f_in5d.xx, gf1.xx,
        mass, m5, ls, t_wick_phase_factor_vec, is_dagger, stop_rsd, max_num_iter)
    return f_out5d

@q.timer
def multiply_m_dwf_qed(
        FieldComplexD f_in5d, FieldComplexD gf1,
        cc.RealD mass, cc.RealD m5, cc.Int ls,
        *,
        t_wick_phase_factor_arr=None,
        cc.Bool is_dagger=False,
        ):
    """
    gf1 = q.mk_left_expanded_field(gf)
    """
    cdef FieldComplexD f_out5d = FieldComplexD()
    cdef cc.vector[cc.ComplexD] t_wick_phase_factor_vec = cc.vector[cc.ComplexD]()
    cdef cc.Int t_size
    cdef cc.Int i
    if t_wick_phase_factor_arr is not None:
        t_size = len(t_wick_phase_factor_arr)
        t_wick_phase_factor_vec.resize(t_size)
        for i in range(t_size):
            t_wick_phase_factor_vec[i] = cc.ccpy_d(t_wick_phase_factor_arr[i])
    cc.multiply_m_dwf_qed(f_out5d.xx, f_in5d.xx, gf1.xx, mass, m5, ls, t_wick_phase_factor_vec, is_dagger)
    return f_out5d

###


import numpy as np

use_reference_implementation = False

mpi_comm = None

def set_mpi_comm(comm):
    global mpi_comm
    mpi_comm = comm

def get_mpi_comm():
    global mpi_comm
    if mpi_comm is None:
        from mpi4py import MPI
        mpi_comm = MPI.COMM_WORLD
    return mpi_comm

def bcast_py(x, root=0, comm=None):
    if comm is None:
        comm = get_mpi_comm()
    return comm.bcast(x, root)

class DistArray:

    """
    self.n : total size for the first dimension (the distributed dimension) of the array.
    self.x : np.ndarray for this MPI process (pad zeros if does not divide).
    self.comm : comm used when generate this array
    """

    def __init__(self, *, comm=None):
        if comm is None:
            comm = get_mpi_comm()
        self.x = np.zeros(1, dtype=np.float64)
        self.n = 1
        self.comm = comm

    def __repr__(self):
        return f"DistArray({self.n!r},\n{self.x!r})"

    def sum_ref(self, axis=None, *, keepdims=False):
        """
        return the results of sum, never as DistArray.
        Collective operation.
        """
        v = all_gather_arr(self)
        return v.sum(axis=axis, keepdims=keepdims)

    def sum(self, axis=None, *, keepdims=False):
        """
        return the results of sum, never as DistArray.
        Collective operation.
        """
        if use_reference_implementation:
            return self.sum_ref(axis=axis, keepdims=keepdims)
        comm = self.comm
        if axis is not None:
            if isinstance(axis, int):
                axis = (axis,)
            if 0 not in axis:
                d_v = DistArray(comm=comm)
                d_v.comm = comm
                d_v.n = self.n
                d_v.x = self.x.sum(axis=axis, keepdims=keepdims)
                return all_gather_arr(d_v)
        v = self.x.sum(axis=axis, keepdims=True)
        comm.Allreduce(v.copy(), v)
        return v.sum(axis=axis, keepdims=keepdims)

    def __add__(self, v):
        comm = self.comm
        d_ret = DistArray(comm=comm)
        d_ret.n = self.n
        if isinstance(v, DistArray):
            assert self.comm is v.comm
            assert self.n == v.n
            assert len(self.x.shape) == len(v.x.shape)
            d_ret.x = self.x + v.x
        else:
            if isinstance(v, np.ndarray):
                assert len(self.x.shape) >= len(v.shape)
            d_ret.x = self.x + v
        return d_ret

    def __sub__(self, v):
        comm = self.comm
        d_ret = DistArray(comm=comm)
        d_ret.n = self.n
        if isinstance(v, DistArray):
            assert self.comm is v.comm
            assert self.n == v.n
            assert len(self.x.shape) == len(v.x.shape)
            d_ret.x = self.x - v.x
        else:
            if isinstance(v, np.ndarray):
                assert len(self.x.shape) >= len(v.shape)
            d_ret.x = self.x - v
        return d_ret

    def __mul__(self, v):
        comm = self.comm
        d_ret = DistArray(comm=comm)
        d_ret.n = self.n
        if isinstance(v, DistArray):
            assert self.comm is v.comm
            assert self.n == v.n
            assert len(self.x.shape) == len(v.x.shape)
            d_ret.x = self.x * v.x
        else:
            if isinstance(v, np.ndarray):
                assert len(self.x.shape) >= len(v.shape)
            d_ret.x = self.x * v
        return d_ret

    def __truediv__(self, v):
        comm = self.comm
        d_ret = DistArray(comm=comm)
        d_ret.n = self.n
        if isinstance(v, DistArray):
            assert self.comm is v.comm
            assert self.n == v.n
            assert len(self.x.shape) == len(v.x.shape)
            d_ret.x = self.x / v.x
        else:
            if isinstance(v, np.ndarray):
                assert len(self.x.shape) >= len(v.shape)
            d_ret.x = self.x / v
        return d_ret

    __radd__ = __add__

    __rmul__ = __mul__

    def __rsub__(self, v):
        comm = self.comm
        d_ret = DistArray(comm=comm)
        d_ret.n = self.n
        if isinstance(v, DistArray):
            assert self.comm is v.comm
            assert self.n == v.n
            assert len(self.x.shape) == len(v.x.shape)
            d_ret.x = v.x - self.x
        else:
            if isinstance(v, np.ndarray):
                assert len(self.x.shape) >= len(v.shape)
            d_ret.x = v - self.x
        return d_ret

    def __rtruediv__(self, v):
        comm = self.comm
        d_ret = DistArray(comm=comm)
        d_ret.n = self.n
        if isinstance(v, DistArray):
            assert self.comm is v.comm
            assert self.n == v.n
            assert len(self.x.shape) == len(v.x.shape)
            d_ret.x = v.x / self.x
        else:
            if isinstance(v, np.ndarray):
                assert len(self.x.shape) >= len(v.shape)
            d_ret.x = v / self.x
        return d_ret

    def __matmul__(self, v):
        return d_matmul(self, v)

    def transpose2d_ref(self):
        """
        return DistArray
        transpose the first two dims
        Collective operation.
        """
        assert len(self.x.shape) >= 2
        comm = self.comm
        rank = comm.Get_rank()
        root = 0
        axes = list(range(len(self.x.shape)))
        axes[:2] = [ 1, 0, ]
        v = gather_arr(self, root)
        if rank == root:
            v = v.transpose(axes)
        d_v = scatter_arr(v, root, self.comm)
        return d_v

    def transpose2d(self):
        """
        return DistArray
        transpose the first two dims
        Collective operation.
        """
        if use_reference_implementation:
            return self.transpose2d_ref()
        comm = self.comm
        size = comm.Get_size()
        axes = list(range(len(self.x.shape)))
        axes[:2] = [ 1, 0, ]
        vec = self.x.transpose(axes)
        vec_len = len(vec)
        d_vec_len = (vec_len - 1) // size + 1
        new_vec_len = d_vec_len * size
        assert new_vec_len >= vec_len
        if new_vec_len > vec_len:
            # pad zeros to vec
            new_vec = np.empty((new_vec_len,) + vec.shape[1:], dtype=vec.dtype)
            new_vec[:vec_len] = vec
            new_vec[vec_len:] = 0
            vec = new_vec
        else:
            # make sure it is contiguous
            assert d_vec_len * size == vec_len
        vec = np.ascontiguousarray(vec)
        nvec = np.empty((size, d_vec_len,) + vec.shape[1:], dtype=vec.dtype)
        comm.Alltoall(vec, nvec)
        axes = list(range(len(nvec.shape)))
        axes[:2] = [ 1, 0, ]
        nvec = nvec.transpose(axes)
        nvec = nvec.reshape(d_vec_len, size * vec.shape[1], *vec.shape[2:])
        assert nvec.shape[1] >= self.n
        nvec = nvec[:, :self.n]
        nvec = np.ascontiguousarray(nvec)
        d_ret = DistArray(comm=comm)
        d_ret.n = vec_len
        d_ret.x = nvec
        return d_ret

    def transpose(self, axes=None):
        shape = self.x.shape
        assert len(shape) >= 1
        if len(shape) == 1:
            return self
        elif len(shape) == 2:
            if (axes is None) or (tuple(axes) == (1, 0,)):
                return self.transpose2d()
            else:
                assert tuple(axes) == (0, 1,)
                return self
        else:
            raise Exception(f"DistArray.transpose: axes={axes}")

    def conj(self):
        """
        return DistArray
        Collective operation.
        """
        comm = self.comm
        d_ret = DistArray(comm=comm)
        d_ret.n = self.n
        d_ret.x = self.x.conj()
        return d_ret

###

def d_matmul_ref(d_mat, d_vec):
    """
    return DistArray
    d_mat is distributed mat
    d_vec is distributed vec
    """
    assert d_mat.comm is d_vec.comm
    assert len(d_mat.x.shape) >= 2
    comm = d_mat.comm
    rank = comm.Get_rank()
    root = 0
    mat = gather_arr(d_mat, root)
    vec = gather_arr(d_vec, root)
    if root == rank:
        ret = np.matmul(mat, vec)
    else:
        ret = None
    d_ret = scatter_arr(ret, root, comm)
    return d_ret

def d_matmul(d_mat, d_vec):
    """
    return DistArray
    d_mat is distributed mat
    d_vec is distributed vec
    """
    if use_reference_implementation:
        return d_matmul_ref(d_mat, d_vec)
    assert d_mat.comm is d_vec.comm
    assert len(d_mat.x.shape) >= 2
    comm = d_mat.comm
    d_ret = DistArray(comm=comm)
    vec = all_gather_arr(d_vec)
    ret = np.matmul(d_mat.x, vec)
    ret = np.ascontiguousarray(ret)
    d_ret.n = d_mat.n
    d_ret.x = ret
    return d_ret

def d_trace_ref(d_mat):
    assert len(d_mat.x.shape) >= 2
    mat = all_gather_arr(d_mat)
    return np.trace(mat)

def d_trace(d_mat):
    if use_reference_implementation:
        return d_trace_ref(d_mat)
    assert len(d_mat.x.shape) >= 2
    comm = d_mat.comm
    size = comm.Get_size()
    rank = comm.Get_rank()
    d_vec_len = len(d_mat.x)
    r = np.trace(d_mat.x, offset=rank * d_vec_len)
    if isinstance(r, np.ndarray):
        comm.Allreduce(r.copy(), r)
    else:
        r = comm.allreduce(r)
    return r

###

def scatter_arr(vec, root=0, comm=None):
    """
    return DistArray
    only need value of vec on the root node
    Pad zeros to vec if len(vec) % comm.Get_size() != 0
    """
    if comm is None:
        comm = get_mpi_comm()
    d_vec = DistArray(comm=comm)
    size = comm.Get_size()
    rank = comm.Get_rank()
    if rank == root:
        assert isinstance(vec, np.ndarray)
        shape_dtype = (vec.shape, vec.dtype,)
    else:
        shape_dtype = None
    shape, dtype = comm.bcast(shape_dtype, root)
    assert len(shape) >= 1
    vec_len = shape[0]
    d_vec_len = (vec_len - 1) // size + 1
    new_vec_len = d_vec_len * size
    assert new_vec_len >= vec_len
    if rank == root:
        if new_vec_len > vec_len:
            # pad zeros to vec
            new_vec = np.empty((new_vec_len,) + shape[1:], dtype=dtype)
            new_vec[:vec_len] = vec
            new_vec[vec_len:] = 0
            vec = new_vec
        # make sure vec is contiguous
        vec = np.ascontiguousarray(vec)
    d_vec.n = vec_len
    d_vec.x = np.empty((d_vec_len,) + shape[1:], dtype=dtype)
    comm.Scatter(vec, d_vec.x, root)
    d_vec.comm = comm
    return d_vec

def gather_arr(d_vec, root=0):
    """
    return np.ndarray on node root
    return None on other nodes
    collective operation, need to run on all nodes.
    Remove padded zeros added
    """
    assert isinstance(d_vec, DistArray)
    comm = d_vec.comm
    size = comm.Get_size()
    rank = comm.Get_rank()
    d_vec_len = len(d_vec.x)
    new_vec_len = d_vec_len * size
    if rank == root:
        vec_g = np.zeros((new_vec_len,) + d_vec.x.shape[1:], dtype=d_vec.x.dtype)
    else:
        vec_g = None
    d_vec.x = np.ascontiguousarray(d_vec.x)
    comm.Gather(d_vec.x, vec_g, root)
    if rank == root:
        assert new_vec_len >= d_vec.n
        vec_g = vec_g[:d_vec.n]
        # vec_g = np.ascontiguousarray(vec_g)
    return vec_g

def all_gather_arr(d_vec):
    """
    return np.ndarray on all nodes
    Remove padded zeros
    """
    assert isinstance(d_vec, DistArray)
    comm = d_vec.comm
    size = comm.Get_size()
    d_vec_len = len(d_vec.x)
    new_vec_len = d_vec_len * size
    vec_g = np.zeros((new_vec_len,) + d_vec.x.shape[1:], dtype=d_vec.x.dtype)
    d_vec.x = np.ascontiguousarray(d_vec.x)
    comm.Allgather(d_vec.x, vec_g)
    assert new_vec_len >= d_vec.n
    vec_g = vec_g[:d_vec.n]
    # vec_g = np.ascontiguousarray(vec_g)
    return vec_g



from .rbc_ukqcd_params import set_param, get_param

tag = "a_inv_gev"
set_param("test-4nt16", tag, value=1.0)
set_param("24D", tag, value=1.023)
set_param("32D", tag, value=1.023)
set_param("48D", tag, value=1.023)
set_param("48I", tag, value=1.730)
set_param("64I", tag, value=2.359)
set_param("64I-pq", tag, value=2.359)
set_param("32IfineH", tag, value=3.148)

tag = "m_pi"
set_param("test-4nt16", tag, value=0.2)
set_param("24D", tag, value=0.13975)
set_param("32D", tag, value=0.13975)
set_param("48D", tag, value=0.13975)
set_param("48I", tag, value=0.08049)
set_param("32IfineH", tag, value=0.11790)

tag = "m_kk"
set_param("test-4nt16", tag, value=0.4)
set_param("24D", tag, value=0.504154)
set_param("32D", tag, value=0.504154)
set_param("48D", tag, value=0.504154)
set_param("48I", tag, value=0.28853)
set_param("32IfineH", tag, value=0.17720)

tag = "zz_vv"
set_param("test-4nt16", tag, value=0.7)
set_param("24D", tag, value=0.72672)
set_param("32D", tag, value=0.72672)
set_param("48D", tag, value=0.72672)
set_param("48I", tag, value=0.71076)
set_param("64I", tag, value=0.74293)
set_param("64I-pq", tag, value=0.74293)
set_param("32IfineH", tag, value=0.77700)

tag = "zz_aa"
set_param("test-4nt16", tag, value=0.7)
set_param("24D", tag, value=0.73457)
set_param("32D", tag, value=0.73457)
set_param("48D", tag, value=0.73457)
set_param("48I", tag, value=0.71191)
set_param("64I", tag, value=0.74341)
set_param("64I-pq", tag, value=0.74341)
set_param("32IfineH", tag, value=0.77779)

tag = "m_res"
set_param("test-4nt16", tag, value=0.001)
set_param("24D", tag, value=0.0022824)
set_param("32D", tag, value=0.0022824)
set_param("48D", tag, value=0.0022824)
set_param("48I", tag, value=0.0006102)
set_param("64I", tag, value=0.0003116)
set_param("64I-pq", tag, value=0.0003116)
set_param("32IfineH", tag, value=0.0006296)

tag = "m_l"
set_param("test-4nt8", tag, value=0.01)
set_param("test-4nt16", tag, value=0.01)
set_param("24D", tag, value=0.00107)
set_param("32D", tag, value=0.00107)
set_param("48D", tag, value=0.00107)
set_param("48I", tag, value=0.00078)
set_param("64I", tag, value=0.000678)
set_param("64I-pq", tag, value=0.0006203)
set_param("32IfineH", tag, value=0.0047)

tag = "m_h"
set_param("test-4nt8", tag, value=0.04)
set_param("test-4nt16", tag, value=0.04)
set_param("24D", tag, value=0.0850)
set_param("32D", tag, value=0.0850)
set_param("48D", tag, value=0.0850)
set_param("48I", tag, value=0.0362)
set_param("64I", tag, value=0.02661)
set_param("64I-pq", tag, value=0.02539)
set_param("32IfineH", tag, value=0.0186)

# PHYSICAL REVIEW D 93, 074505 (2016)
# zz_m_l * m_l => m_l in MSbar scheme 3 GeV
tag = "zz_m_l"
set_param("64I", tag, value=2.997 / 2.198)
set_param("64I-pq", tag, value=2.997 / 2.198)
set_param("48I", tag, value=2.997 / 2.198 * 0.9715)

# PHYSICAL REVIEW D 93, 074505 (2016)
# zz_m_h * m_h => m_l in MSbar scheme 3 GeV
tag = "zz_m_h"
set_param("64I", tag, value=81.64 / 60.62)
set_param("64I-pq", tag, value=81.64 / 60.62)
set_param("48I", tag, value=81.64 / 60.62 * 0.9628)

tag = "zz_ss_l"
set_param("64I", tag, value=1 / get_param("64I", "zz_m_l"))
set_param("64I-pq", tag, value=1 / get_param("64I-pq", "zz_m_l"))
set_param("48I", tag, value=1 / get_param("48I", "zz_m_l"))

tag = "zz_ss_h"
set_param("64I", tag, value=1 / get_param("64I", "zz_m_h"))
set_param("64I-pq", tag, value=1 / get_param("64I-pq", "zz_m_h"))
set_param("48I", tag, value=1 / get_param("48I", "zz_m_h"))



# cython: binding=True, embedsignature=True, c_string_type=unicode, c_string_encoding=utf8

cimport qlat_utils.everything
from qlat_utils.all cimport *
from . cimport everything as cc
from cpython cimport Py_buffer
from cpython.buffer cimport PyBUF_FORMAT

import qlat_utils as q
import numpy as np

### -------------------------------------------------------------------

def mpi_level_count():
    return cc.mpi_level_count()

def begin(int id_node, Coordinate size_node, int color=0):
    cc.begin(id_node, size_node.xx, color)

def end(cc.Bool is_preserving_cache=False):
    if not is_preserving_cache:
        q.clean_cache()
    cc.end(is_preserving_cache)
    if mpi_level_count() == 0:
        q.set_display_method()

### -------------------------------------------------------------------

def get_size_node():
    cdef Coordinate x = Coordinate()
    # cc.assign_direct(x.xx, cc.get_size_node())
    x.xx = cc.get_size_node()
    return x

def get_coor_node():
    cdef Coordinate x = Coordinate()
    # cc.assign_direct(x.xx, cc.get_coor_node())
    x.xx = cc.get_coor_node()
    return x

### -------------------------------------------------------------------

def bcast_long(cc.Long x, int root=0):
    cc.bcast(x, root)
    return x

def bcast_double(double x, int root=0):
    cc.bcast(x, root)
    return x

def bcast_complex(cc.PyComplexD x, int root=0):
    cc.bcast(cc.ccpy_d(x), root)
    return x

def bcast_lat_data_in_place(LatData ld, int root=0):
    cc.bcast(ld.xx, root)
    return ld

def bcast_lat_data(LatData ld, int root=0):
    cdef LatData ld1 = ld.copy()
    return bcast_lat_data_in_place(ld1, root)

### -------------------------------------------------------------------

def get_id_node_list_for_shuffle():
    """
    return ``list``
    ``list`` is a list of ``id_node``s, which are intended to perform IO.
    ``list[id_node_in_shuffle] = id_node``
    """
    cdef list id_node_list = []
    cdef cc.std_vector[cc.Int] id_node_vec = cc.get_id_node_list_for_shuffle()
    for id_node in id_node_vec:
        id_node_list.append(id_node)
    return id_node_list

def get_id_node_in_shuffle_list():
    """
    return ``list``
    ``list[id_node] = id_node_in_shuffle``
    Related to ``get_id_node_list_for_shuffle()``
    """
    cdef list id_node_in_shuffle_list = []
    cdef cc.std_vector[cc.Int] id_node_in_shuffle_vec = cc.get_id_node_in_shuffle_list()
    for id_node in id_node_in_shuffle_vec:
        id_node_in_shuffle_list.append(id_node)
    return id_node_in_shuffle_list

### -------------------------------------------------------------------

def glb_sum_long(cc.Long x):
    cc.glb_sum(x)
    return x

def glb_sum_double(cc.RealD x):
    cc.glb_sum(x)
    return x

def glb_sum_complex(cc.PyComplexD x):
    cdef cc.ComplexD xx = cc.ccpy_d(x)
    cc.glb_sum(xx)
    x = cc.pycc_d(xx)
    return x

def glb_sum_lat_data_in_place(LatData ld):
    cc.glb_sum(ld.xx)
    return ld

def glb_sum_lat_data(LatData ld):
    cdef LatData ld1 = ld.copy()
    return glb_sum_lat_data_in_place(ld1)

@q.timer
def glb_sum_np(x):
    """
    x does NOT change
    """
    shape = x.shape
    dtype = x.dtype
    l = list(x.flatten())
    ld = LatData()
    if dtype == np.dtype('float64'):
        ld.from_list(l, is_complex=False)
    elif dtype == np.dtype('int64'):
        ld.from_list(list(map(float, l)), is_complex=False)
    elif dtype == np.dtype('complex128'):
        ld.from_list(l, is_complex=True)
    else:
        q.displayln(dtype)
        assert False
    ld.glb_sum_in_place()
    return np.array(ld.to_list(), dtype=dtype).reshape(shape)

@q.timer
def glb_sum(x):
    """
    x does NOT change
    """
    if isinstance(x, float):
        return glb_sum_double(x)
    elif isinstance(x, complex):
        return glb_sum_complex(x)
    elif isinstance(x, (int, np.int64)):
        return glb_sum_long(x)
    elif isinstance(x, np.ndarray):
        return glb_sum_np(x)
    elif isinstance(x, list):
        return [ glb_sum(x_i) for x_i in x ]
    elif isinstance(x, tuple):
        return tuple(glb_sum(x_i) for x_i in x)
    else:
        # possible types: q.LatData
        return x.glb_sum()

### -------------------------------------------------------------------



from . cimport everything as cqlat
from .field_types cimport FieldWilsonMatrix, FieldWilsonVector, FieldSpinMatrix
from .selected_field_types cimport SelectedFieldWilsonMatrix, SelectedFieldSpinMatrix
from .selected_points_types cimport SelectedPointsWilsonMatrix, SelectedPointsSpinMatrix

cdef class Prop(FieldWilsonMatrix):

    cdef cqlat.Handle[cqlat.Prop] xxx(self)

cdef class SelProp(SelectedFieldWilsonMatrix):

    cdef cqlat.Handle[cqlat.SelProp] xxx(self)

cdef class PselProp(SelectedPointsWilsonMatrix):

    cdef cqlat.Handle[cqlat.PselProp] xxx(self)

cdef class FermionField4d(FieldWilsonVector):

    cdef cqlat.Handle[cqlat.FermionField4d] xxx(self)

cdef class SpinProp(FieldSpinMatrix):

    cdef cqlat.Handle[cqlat.SpinProp] xxx(self)


import sys
import qlat as q
import numpy as np
from pprint import pformat

size_node_list = [
        [1, 1, 1, 1],
        [1, 1, 1, 2],
        [1, 1, 1, 3],
        [1, 1, 1, 4],
        [1, 1, 1, 6],
        [1, 1, 1, 8],
        [1, 2, 2, 4],
        [2, 2, 2, 4],
        [2, 2, 2, 4],
        ]

q.begin_with_mpi()

if len(sys.argv) == 1:
    q.displayln_info("Usage: topo-measure [ --source source_config ] [ --output path ] [ --density-field ] [ --show-topo-terms ]")

q.displayln_info("Topological charge measurement with Qlattice")
q.displayln_info("by Luchang Jin")
q.displayln_info("2024/01/25")

p_source = q.get_arg("--source")
p_output = q.get_arg("--output")
is_density_field = q.get_option("--density-field")
is_show_topo_terms = q.get_option("--show-topo-terms")

info_path = p_output

if is_density_field:
    assert p_output is not None
    density_field_path=info_path
else:
    density_field_path=None

def load():
    if p_source is None:
        q.displayln_info("Need to provide source file with '--source filename'. Use a sample gauge field for now.")
        total_site = q.Coordinate([ 4, 4, 4, 8, ])
        geo = q.Geometry(total_site)
        gf = q.GaugeField(geo)
        rs = q.RngState("seed")
        gf.set_rand(rs.split("gf-init"), 0.5, 10)
    else:
        gf = q.GaugeField()
        gf.load(p_source)
    q.clear_all_caches()
    q.clear_mem_cache()
    return gf

gf = load()

topo_list, energy_list, = q.smear_measure_topo(
        gf,
        info_path=info_path,
        density_field_path=density_field_path,
        is_show_topo_terms=is_show_topo_terms,
        )

if info_path is None:
    q.displayln_info("To save the result, use '--output path'. Print to screen for now.")
    q.displayln_info(pformat(topo_list))
    q.displayln_info(pformat(energy_list))

q.timer_display()

q.end_with_mpi()

exit()



# cython: binding=True, embedsignature=True, c_string_type=unicode, c_string_encoding=utf8

from qlat_utils.all cimport *
from . cimport everything as cc
from cpython cimport Py_buffer
from cpython.buffer cimport PyBUF_FORMAT

import qlat_utils as q
import numpy as np

### -------------------------------------------------------------------

cdef class Geometry:

    def __cinit__(self):
        self.xx = cc.Geometry()
        self.cdata = <cc.Long>&(self.xx)

    def __init__(self, *args):
        """
        Geometry()
        Geometry(total_site)
        Geometry(id_node, size_node, node_site)
        Geometry(coor_node, size_node, node_site)
        """
        n_args = len(args)
        if n_args == 0:
            self.xx.init()
        elif n_args == 1:
            if isinstance(args[0], Coordinate):
                self.init_from_total_site(*args)
            else:
                fname = q.get_fname()
                raise Exception(f"{fname}: args={args}")
        elif n_args == 3:
            if isinstance(args[0], Coordinate):
                self.init_from_coor_node_site(*args)
            elif isinstance(args[0], int):
                self.init_from_id_node_site(*args)
            else:
                fname = q.get_fname()
                raise Exception(f"{fname}: args={args}")
        else:
            fname = q.get_fname()
            raise Exception(f"{fname}: args={args}")

    def init_from_total_site(self, Coordinate total_site):
        """
        Use `total_site` to initialize `Geometry`.
        Will use the global `geon` to determine `id_node` and `size_node`.
        """
        self.xx.init(total_site.xx)

    def init_from_id_node_site(self, int id_node, Coordinate size_node, Coordinate node_site):
        """
        initialize Geometry without call MPI functions.
        #
        corr_node = q.coordinate_from_index(id_node, size_node)
        id_node = q.index_from_coordinate(corr_node, size_node)
        #
        total_site = size_node * node_site
        """
        self.xx.init(id_node, size_node.xx, node_site.xx)

    def init_from_coor_node_site(self, Coordinate coor_node, Coordinate size_node, Coordinate node_site):
        """
        initialize Geometry without call MPI functions.
        #
        corr_node = q.coordinate_from_index(id_node, size_node)
        id_node = q.index_from_coordinate(corr_node, size_node)
        #
        total_site = size_node * node_site
        """
        self.xx.init(coor_node.xx, size_node.xx, node_site.xx)

    def __imatmul__(self, Geometry v1):
        cc.assign_direct(self.xx, v1.xx)
        return self

    def copy(self, is_copying_data=True):
        x = Geometry()
        if is_copying_data:
            x @= self
        return x

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def __eq__(self, Geometry other):
        return self.xx == other.xx

    @property
    def total_site(self):
        cdef Coordinate x = Coordinate()
        cc.assign_direct(x.xx, self.xx.total_site())
        return x

    @property
    def total_volume(self):
        return self.xx.total_volume()

    @property
    def spatial_volume(self):
        return self.xx.spatial_volume()

    @property
    def local_site(self):
        cdef Coordinate x = Coordinate()
        cc.assign_direct(x.xx, self.xx.local_site())
        return x

    @property
    def local_volume(self):
        return self.xx.local_volume()

    @property
    def local_volume_expanded(self):
        return self.xx.local_volume_expanded()

    @property
    def eo(self):
        return self.xx.eo

    @property
    def expansion_left(self):
        cdef Coordinate x = Coordinate()
        cc.assign_direct(x.xx, self.xx.expansion_left)
        return x

    @property
    def expansion_right(self):
        cdef Coordinate x = Coordinate()
        cc.assign_direct(x.xx, self.xx.expansion_right)
        return x

    @property
    def id_node(self):
        return self.xx.geon.id_node

    @property
    def num_node(self):
        return self.xx.geon.num_node

    @property
    def coor_node(self):
        cdef Coordinate x = Coordinate()
        cc.assign_direct(x.xx, self.xx.geon.coor_node)
        return x

    @property
    def size_node(self):
        cdef Coordinate x = Coordinate()
        cc.assign_direct(x.xx, self.xx.geon.size_node)
        return x

    @property
    def node_site(self):
        cdef Coordinate x = Coordinate()
        cc.assign_direct(x.xx, self.xx.node_site)
        return x

    @property
    def is_only_local(self):
        return self.xx.is_only_local

    def __str__(self):
        return self.show()

    def __repr__(self):
        return self.show()

    def show(self):
        cdef Coordinate total_site = self.total_site
        cdef Coordinate expan_left = self.expansion_left
        cdef Coordinate expan_right = self.expansion_right
        cdef int eo = self.eo
        cdef Coordinate zero = Coordinate()
        if expan_left == zero and expan_right == zero and eo == 0:
            return f"Geometry({total_site.to_list()})"
        else:
            return f"Geometry({total_site.to_list()}, expan_left={expan_left.to_list()}, expan_right={expan_right.to_list()}, eo={eo})"

    def coordinate_g_from_l(self, Coordinate xl not None):
        cdef Coordinate xg = Coordinate()
        cc.assign_direct(xg.xx, self.xx.coordinate_g_from_l(xl.xx))
        return xg

    def coordinate_l_from_g(self, Coordinate xg not None):
        cdef Coordinate xl = Coordinate()
        cc.assign_direct(xl.xx, self.xx.coordinate_l_from_g(xg.xx))
        return xl

    def index_from_coordinate(self, Coordinate xl not None):
        return self.xx.index_from_coordinate(xl.xx)

    def coordinate_from_index(self, cc.Long index):
        cdef Coordinate xl = Coordinate()
        cc.assign_direct(xl.xx, self.xx.coordinate_from_index(index))
        return xl

    def index_from_g_coordinate(self, Coordinate xg not None):
        return self.xx.index_from_g_coordinate(xg.xx)

    def g_coordinate_from_index(self, cc.Long index):
        cdef Coordinate xg = Coordinate()
        cc.assign_direct(xg.xx, self.xx.g_coordinate_from_index(index))
        return xg

    def is_local(self, Coordinate xl not None):
        return self.xx.is_local(xl.xx)

    def is_local_xg(self, Coordinate xg not None):
        """
        return a global coordinate is inside the local volume or not
        """
        cdef cc.Coordinate xl_xx = self.xx.coordinate_l_from_g(xg.xx)
        return self.xx.is_local(xl_xx)

    def xg_arr(self):
        """
        return xg for all local sites
        shape = (geo.local_volume, 4,)
        """
        from .field_selection import mk_xg_field
        f_xg = mk_xg_field(self)
        xg_arr = np.asarray(f_xg, dtype=np.int32)
        cdef cc.Long local_volume = self.local_volume
        return xg_arr.reshape((local_volume, 4,))

    def __getstate__(self):
        """
        State of `Geometry` is different on different nodes.
        """
        size_node = self.size_node
        coor_node = self.coor_node
        node_site = self.node_site
        expan_left = self.expansion_left
        expan_right = self.expansion_right
        return [ size_node, coor_node, node_site, expan_left, expan_right, ]

    def __setstate__(self, state):
        """
        State of `Geometry` is different on different nodes.
        """
        self.__init__()
        cdef Coordinate size_node
        cdef Coordinate coor_node
        cdef Coordinate node_site
        cdef Coordinate expan_left
        cdef Coordinate expan_right
        [ size_node, coor_node, node_site, expan_left, expan_right, ] = state
        self.xx.init(coor_node.xx, size_node.xx, node_site.xx)
        self.xx.expansion_left = expan_left.xx
        self.xx.expansion_right = expan_right.xx
        self.xx.reset_node_site_expanded()

### -------------------------------------------------------------------

def geo_resize(Geometry geo, expansion_left=None, expansion_right=None):
    """
    expansion_left can be None, or int, or Coordinate
    expansion_right can be None, or int, or Coordinate
    Default is zero.
    """
    cdef Coordinate el, er
    if expansion_left is None:
        el = Coordinate()
    elif isinstance(expansion_left, int):
        e = expansion_left
        el = Coordinate([ e, e, e, e, ])
    elif isinstance(expansion_left, list):
        el = Coordinate(expansion_left)
    else:
        assert isinstance(expansion_left, Coordinate)
        el = <Coordinate>expansion_left
    if expansion_right is None:
        er = Coordinate()
    elif isinstance(expansion_right, int):
        e = expansion_right
        er = Coordinate([ e, e, e, e, ])
    elif isinstance(expansion_right, list):
        er = Coordinate(expansion_right)
    else:
        assert isinstance(expansion_right, Coordinate)
        er = <Coordinate>expansion_right
    cdef Geometry geo_new = Geometry()
    cc.assign_direct(geo_new.xx, cc.geo_resize(geo.xx, el.xx, er.xx))
    return geo_new

def geo_eo(Geometry geo, int eo=0):
    cdef Geometry geo_new = Geometry()
    cc.assign_direct(geo_new.xx, cc.geo_eo(geo.xx, eo))
    return geo_new

### -------------------------------------------------------------------



"""
Qlattice main package.\n
Usage::\n
    import qlat as q\n
Will also load ``qlat_utils``.
"""

from qlat_utils import *

set_verbose_level(-1)

from .c import *

from .mpi_utils import *

from .scalar_action import *

from .qm_action import *

from .fthmc import *

from .hmc_stats import *

from .contract_pion import *

from .contract_field import *

from .inverter import *

from .field_analysis import *

from .mat_mpi import *

from .psel_split import *

from .smear_prop import *

from .instanton_map import *

from .flow_scale import *

from . import field_double

set_verbose_level()



# cython: binding=True, embedsignature=True, c_string_type=unicode, c_string_encoding=utf8

from qlat_utils.all cimport *
from . cimport everything as cc
from .geometry cimport Geometry
from .field_base cimport (
        FieldBase,
        SelectedFieldBase,
        SelectedPointsBase,
        )
from .field_types cimport (
        FieldInt8t,
        FieldRealD,
        FieldChar,
        )
from .selected_field_types cimport SelectedFieldRealD
from .selected_points_types cimport SelectedPointsRealD

from cpython cimport Py_buffer
from cpython.buffer cimport PyBUF_FORMAT

import cqlat as c
import qlat_utils as q
import numpy as np
import math

from .geometry import geo_resize
from .field_base import (
        Field,
        SelectedField,
        SelectedPoints,
        )

def field_expanded(f, expansion_left, expansion_right):
    geo = f.geo
    multiplicity = f.multiplicity
    geo_e = geo_resize(geo, expansion_left, expansion_right)
    f_e = type(f)(geo_e, multiplicity)
    f_e @= f
    return f_e

def refresh_expanded(field, comm_plan=None):
    if comm_plan is None:
        return c.refresh_expanded_field(field)
    else:
        return c.refresh_expanded_field(field, comm_plan)

def refresh_expanded_1(field):
    return c.refresh_expanded_1_field(field)


cdef class CommMarks(FieldInt8t):

    def __init__(self, Geometry geo=None, int multiplicity=0):
        super().__init__(geo, multiplicity)

###

cdef class CommPlan:

    def __cinit__(self):
        self.cdata = <cc.Long>&(self.xx)

    def __imatmul__(self, CommPlan v1):
        self.xx = v1.xx
        return self

    def copy(self, is_copying_data=True):
        x = type(self)()
        if is_copying_data:
            x @= self
        return x

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

###

def make_field_expand_comm_plan(CommMarks comm_marks):
    """
    comm_marks is of type Field(ElemTypeInt8t)
    """
    cp = CommPlan()
    c.make_field_expand_comm_plan(cp, comm_marks)
    return cp

def mk_phase_field(Geometry geo, lmom):
    """
    lmom is in lattice momentum unit
    exp(i * 2*pi/L * lmom \cdot xg )
    """
    f = Field(ElemTypeComplexD, geo, 1)
    c.set_phase_field(f, lmom)
    return f

class FastFourierTransform:

    def __init__(self, fft_infos, *, is_normalizing=False, mode_fft=1):
        # mode_fft in [ 0, 1, ]
        # fft_infos = [ ( fft_dir, is_forward, ), ... ]
        self.fft_infos = fft_infos
        self.is_normalizing = is_normalizing
        self.mode_fft = mode_fft

    def copy(self):
        return self.__copy__()

    def __mul__(self, fields):
        if isinstance(fields, FieldBase):
            return (self * [ fields, ])[0]
        assert isinstance(fields, list)
        for f in fields:
            assert isinstance(f, FieldBase)
        fields = [ f.copy() for f in fields ]
        fft_dirs, fft_is_forwards = zip(*self.fft_infos)
        c.fft_fields(fields, fft_dirs, fft_is_forwards, self.mode_fft)
        if self.is_normalizing and self.fft_infos:
            for field in fields:
                total_site = field.total_site
                scale_factor = 1
                for fft_dir, is_forward in self.fft_infos:
                    scale_factor *= total_site[fft_dir]
                scale_factor = 1.0 / math.sqrt(scale_factor)
                field *= scale_factor
        return fields

###

@q.timer
def mk_fft(is_forward, *, is_only_spatial=False, is_normalizing=False, mode_fft=1):
    if is_only_spatial:
        fft_infos = [
                (0, is_forward,),
                (1, is_forward,),
                (2, is_forward,),
                ]
        return FastFourierTransform(fft_infos, is_normalizing=is_normalizing, mode_fft=mode_fft)
    else:
        fft_infos = [
                (0, is_forward,),
                (1, is_forward,),
                (2, is_forward,),
                (3, is_forward,),
                ]
        return FastFourierTransform(fft_infos, is_normalizing=is_normalizing, mode_fft=mode_fft)

###

@q.timer
def qnorm_field(f):
    if isinstance(f, (FieldBase, SelectedFieldBase, SelectedPointsBase,)):
        f_n = f.qnorm_field()
    else:
        q.displayln_info("qnorm_field:", type(f))
        assert False
    return f_n

@q.timer
def sqrt_selected_points_real_d(SelectedPointsRealD f):
    cdef SelectedPointsRealD f_ret = f.copy(is_copying_data=False)
    cc.set_sqrt_field(f_ret.xx, f.xx)
    return f_ret

@q.timer
def sqrt_selected_field_real_d(SelectedFieldRealD f):
    cdef SelectedFieldRealD f_ret = f.copy(is_copying_data=False)
    cc.set_sqrt_field(f_ret.xx, f.xx)
    return f_ret

@q.timer
def sqrt_field_real_d(FieldRealD f):
    cdef FieldRealD f_ret = f.copy(is_copying_data=False)
    cc.set_sqrt_field(f_ret.xx, f.xx)
    return f_ret

@q.timer
def sqrt_field(f):
    if isinstance(f, FieldRealD):
        f_ret = sqrt_field_real_d(f)
    elif isinstance(f, SelectedFieldRealD):
        f_ret = sqrt_selected_field_real_d(f)
    elif isinstance(f, SelectedPointsRealD):
        f_ret = sqrt_selected_points_real_d(f)
    else:
        q.displayln_info("sqrt_field:", type(f))
        assert False
    return f_ret

###

@q.timer
def field_char_shift(FieldChar f, Coordinate shift):
    """
    return `sf`
    `sf` is the new shifted Field.
    `f` is not changed.
    `shift` is the coordinate to shift the field (or None, which means no shift)
    roughly shifted_field[(xg + shift) % total_site] == original_field[xg]
    """
    cdef FieldChar sf = FieldChar()
    cc.field_shift(sf.xx, f.xx, shift.xx)
    return sf

@q.timer
def field_shift(FieldBase f, Coordinate shift):
    cdef FieldChar fc = FieldChar()
    f.swap_cast(fc)
    fc = field_char_shift(fc, shift)
    sf = type(f)()
    sf.swap_cast(fc)
    return sf

@q.timer
def shuffle_field_char(FieldChar f, Coordinate new_size_node):
    """
    return f_list
    """
    cdef cc.std_vector[cc.Field[cc.Char]] fs
    fs = cc.std_vector[cc.Field[cc.Char]]()
    cc.shuffle_field(fs, f.xx, new_size_node.xx)
    cdef cc.Int num_field = <cc.Int>fs.size()
    cdef cc.Int i
    cdef FieldChar f1
    cdef list f_list = []
    for i in range(num_field):
        f1 = FieldChar()
        cc.qswap(f1.xx, fs[i])
        f_list.append(f1)
    return f_list

@q.timer
def shuffle_field_char_back(FieldChar fc, list f_list, Coordinate new_size_node):
    """
    Modify `fc` in place.
    NOTE: `fc` needs to have correct size.
    """
    cdef cc.std_vector[cc.Field[cc.Char]] fs
    cdef cc.Int num_field = <cc.Int>len(f_list)
    cdef cc.Int i
    cdef FieldChar f1
    fs = cc.std_vector[cc.Field[cc.Char]](num_field)
    for i in range(num_field):
        f1 = f_list[i]
        cc.qswap(f1.xx, fs[i])
    cc.shuffle_field_back(fc.xx, fs, new_size_node.xx)
    for i in range(num_field):
        f1 = f_list[i]
        cc.qswap(f1.xx, fs[i])

@q.timer
def shuffle_field(FieldBase f, Coordinate new_size_node):
    cdef FieldChar fc = FieldChar()
    f.swap_cast(fc)
    cdef list fc_list = shuffle_field_char(fc, new_size_node)
    f.swap_cast(fc)
    cdef list f_list = []
    for i in range(len(fc_list)):
        f1 = type(f)()
        f1.swap_cast(fc_list[i])
        f_list.append(f1)
    return f_list

@q.timer
def shuffle_field_back(FieldBase f, list f_list, Coordinate new_size_node):
    """
    Modify `f` in place
    NOTE: `f` needs to have correct size.
    """
    cdef FieldChar f1
    cdef list fc_list = []
    for i in range(len(f_list)):
        f1 = FieldChar()
        f_list[i].swap_cast(f1)
        fc_list.append(f1)
    cdef FieldChar fc = FieldChar()
    f.swap_cast(fc)
    shuffle_field_char_back(fc, fc_list, new_size_node)
    f.swap_cast(fc)
    for i in range(len(fc_list)):
        f_list[i].swap_cast(fc_list[i])


# cython: binding=True, embedsignature=True, c_string_type=unicode, c_string_encoding=utf8

from . cimport everything as cc
from qlat_utils.all cimport *
cimport numpy
import numpy as np
import qlat_utils as q
from .field_selection cimport FieldSelection, PointsSelection
from .field_types cimport FieldRealD, FieldComplexD
from .selected_points_types cimport SelectedPointsRealD, SelectedPointsComplexD, SelectedPointsWilsonMatrix
from .propagator cimport Prop, SelProp, PselProp
from .geometry cimport Geometry

cdef numpy.ndarray sl_arr_from_sl_table(const cc.SlTable& x):
    """
    return sl_arr
    sl_arr.shape == (s_limit, l_limit,)
    sl_arr.dtype == np.complex128
    """
    cdef cc.Long s_limit = x.s_limit
    cdef cc.Long l_limit = x.l_limit
    cdef numpy.ndarray sl_arr = np.zeros((s_limit, l_limit,), dtype=np.complex128)
    cdef cc.PyComplexD[:] ptr = sl_arr.ravel()
    cdef cc.Long table_size = x.table.size()
    cc.memcpy(&ptr[0], x.table.data(), table_size * sizeof(cc.PyComplexD))
    return sl_arr

@q.timer
def mk_m_z_field_tag(
        PointsSelection psel_d,
        Coordinate xg_x,
        Coordinate xg_y,
        const cc.RealD a,
        const cc.Int tag
        ):
    """
    return smf_d
    smf is SelectedPointsRealD() as the muon-line-field of vertex z
    a is the lattice spacing (when muon_mass = 1). In lattice unit one should use muon_mass * a for this parameter
    tag = 0: sub # preferred
    tag = 1: nosub
    """
    cdef Geometry geo = psel_d.geo
    cdef SelectedPointsRealD smf_d = SelectedPointsRealD(psel_d)
    cc.set_m_z_field_tag(smf_d.xx, psel_d.xx, geo.xx, xg_x.xx, xg_y.xx, a, tag)
    return smf_d

def contract_four_pair_labels(list tags):
    return cc.contract_four_pair_labels(tags)

def contract_two_plus_two_pair_labels():
    return cc.contract_two_plus_two_pair_labels()

@q.timer
def mk_local_current_from_props(
        PselProp sprop1,
        PselProp sprop2,
        ):
    """
    return scf
    -<- gamma5 sprop2^+ gamma5 -<- gamma_mu -<- sprop1 -<-
    """
    cdef PointsSelection psel_d = sprop1.psel
    assert len(psel_d) == len(sprop2.psel)
    cdef Geometry geo = psel_d.geo
    assert geo.local_site == sprop2.psel.geo.local_site
    cdef SelectedPointsWilsonMatrix scf = SelectedPointsWilsonMatrix(psel_d)
    cc.set_local_current_from_props(scf.xx, sprop1.xx, sprop2.xx, psel_d.xx, geo.xx)
    return scf

@q.timer
def mk_psel_d_prob_xy(
        SelectedPointsRealD psel_prob,
        SelectedPointsRealD psel_d_prob,
        const cc.Long idx_xg_x,
        const cc.Long idx_xg_y,
        ):
    """
    return prob_pair, psel_d_prob_xy
    """
    cdef PointsSelection psel = psel_prob.psel
    cdef PointsSelection psel_d = psel_d_prob.psel
    cdef Geometry geo = psel.geo
    if geo.local_site != psel_d.geo.local_site:
        raise Exception(f"psel site: {geo.local_site} ; psel_d site: {psel_d.geo.local_site}.")
    cdef SelectedPointsRealD psel_d_prob_xy = SelectedPointsRealD(psel_d)
    cdef cc.RealD prob_pair = cc.set_psel_d_prob_xy(
            psel_d_prob_xy.xx,
            psel.xx,
            psel_prob.xx,
            psel_d.xx,
            psel_d_prob.xx,
            idx_xg_x,
            idx_xg_y,
            )
    return prob_pair, psel_d_prob_xy

cdef class CurrentMoments:

    def __cinit__(self):
        self.xx.init()

    def __init__(self, *args):
        """
        CurrentMoments()
        CurrentMoments(SelectedPointsWilsonMatrix current, SelectedPointsRealD psel_d_prob_xy)
        """
        cdef cc.Int len_args = len(args)
        if len_args > 0:
            if isinstance(args[0], SelectedPointsWilsonMatrix):
                current, psel_d_prob_xy = args
                self.set_from_current(current, psel_d_prob_xy)
            else:
                raise Exception("CurrentMoments.__init__")

    def __imatmul__(self, CurrentMoments v1):
        cc.assign_direct(self.xx, v1.xx)
        return self

    def copy(self, is_copying_data=True):
        x = CurrentMoments()
        if is_copying_data:
            x @= self
        return x

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    @q.timer
    def set_from_current(
            self,
            SelectedPointsWilsonMatrix current,
            SelectedPointsRealD psel_d_prob_xy,
            ):
        cdef CurrentMoments cm = CurrentMoments()
        cdef PointsSelection psel_d = psel_d_prob_xy.psel
        assert len(psel_d) == len(current.psel)
        cdef Geometry geo = psel_d.geo
        cc.set_current_moments_from_current(self.xx, current.xx, psel_d.xx, psel_d_prob_xy.xx, geo.xx)

    @q.timer
    def glb_sum(self):
        """
        Return results. Do NOT modify self.
        """
        cdef CurrentMoments cm = self.copy()
        cc.glb_sum_current_moments(cm.xx)
        return cm

###

@q.timer
def contract_four_pair_no_glb_sum(
        cc.PyComplexD coef,
        SelectedPointsRealD psel_prob,
        SelectedPointsRealD psel_d_prob,
        const cc.Long idx_xg_x,
        const cc.Long idx_xg_y,
        SelectedPointsRealD smf_d,
        SelectedPointsWilsonMatrix sc_xy,
        SelectedPointsWilsonMatrix sc_yx,
        CurrentMoments cm_xy,
        CurrentMoments cm_yx,
        const cc.Int inv_type,
        const cc.std_vector[cc.std_string]& tags,
        const cc.Long r_sq_limit,
        const cc.RealD muon_mass,
        const cc.RealD z_v
        ):
    """
    return lsl_arr
    #
    lsl_arr.shape == (len(labels), s_limit, l_limit,)
    labels = contract_four_pair_labels(tags)
    #
    default coef = 1.0
    inv_type = 0 : light quark
    inv_type = 1 : strange quark
    tags can be [ "ref-far", "ref-center", "ref-close", ]
    #
    glb_sum for SlTable not yet performed
    #
    x, y are two sampled points (elem of `psel`). `psel_prob` factors are already included.
    z, x_op are summed over with in `fsel`. `fsel_prob` factors are already included.
    """
    cdef PointsSelection psel = psel_prob.psel
    cdef PointsSelection psel_d = psel_d_prob.psel
    cdef Geometry geo = psel.geo
    if geo.local_site != psel_d.geo.local_site:
        raise Exception(f"psel site: {geo.local_site} ; psel_d site: {psel_d.geo.local_site}.")
    assert len(psel_d) == len(smf_d.psel)
    assert len(psel_d) == len(sc_xy.psel)
    assert len(psel_d) == len(sc_yx.psel)
    cdef cc.std_vector[cc.SlTable] sl_table_vec = cc.contract_four_pair_no_glb_sum(
            cc.ccpy_d(coef),
            psel.xx,
            psel_d.xx,
            psel_d_prob.xx,
            geo.xx,
            idx_xg_x,
            idx_xg_y,
            smf_d.xx,
            sc_xy.xx,
            sc_yx.xx,
            cm_xy.xx,
            cm_yx.xx,
            inv_type,
            tags,
            r_sq_limit,
            muon_mass,
            z_v,
            )
    cdef cc.Long s = sl_table_vec.size()
    cdef list sl_arr_list = []
    for i in range(s):
        sl_arr = sl_arr_from_sl_table(sl_table_vec[i])
        sl_arr_list.append(sl_arr)
    return np.array(sl_arr_list, dtype=np.complex128)

@q.timer
def contract_two_plus_two_pair_no_glb_sum(
        cc.PyComplexD coef,
        SelectedPointsRealD psel_prob,
        SelectedPointsRealD psel_lps_prob,
        const cc.Long idx_xg_x,
        SelectedPointsComplexD lps_hvp_x,
        SelectedPointsComplexD edl_list_c,
        const cc.Long r_sq_limit,
        const cc.RealD muon_mass,
        const cc.RealD z_v,
        ):
    """
    return n_points_selected, n_points_computed, lsl_arr
    #
    lsl_arr.shape == (len(labels), s_limit, l_limit,)
    labels = contract_two_plus_two_pair_labels()
    #
    hvp point source at x (rho)
    hvp point sink at y (sigma)
    hvp with external loop source at z (lambda)
    hvp with external loop sink summed over (i)
    psel_edl[k] = xg_z
    edl_list[k][i * 4 + lambda]
    -hvp_x.get_elem(xl_y, sigma * 4 + rho)
    #
    glb_sum for SlTable not yet performed
    #
    x is sampled point (elem of `psel`). `psel_prob` factor is already included.
    y is summed over with adaptive sampling. Weight factors is already included.
    z is sampled point (elem of `psel`) and summed over. `psel_prob` factor is already included.
    x_op is summed over all points.
    """
    cdef Geometry geo = psel_prob.psel.geo
    cdef PointsSelection psel = psel_prob.psel
    cdef PointsSelection psel_lps = psel_lps_prob.psel
    cdef cc.Long n_points_selected = 0
    cdef cc.Long n_points_computed = 0
    cdef cc.std_vector[cc.SlTable] sl_table_vec = cc.contract_two_plus_two_pair_no_glb_sum(
            n_points_selected,
            n_points_computed,
            cc.ccpy_d(coef),
            geo.xx,
            psel.xx,
            psel_prob.xx,
            psel_lps.xx,
            psel_lps_prob.xx,
            idx_xg_x,
            lps_hvp_x.xx,
            edl_list_c.xx,
            r_sq_limit,
            muon_mass,
            z_v,
            )
    cdef cc.Long s = sl_table_vec.size()
    cdef list sl_arr_list = []
    for i in range(s):
        sl_arr = sl_arr_from_sl_table(sl_table_vec[i])
        sl_arr_list.append(sl_arr)
    lsl_arr = np.array(sl_arr_list, dtype=np.complex128)
    return n_points_selected, n_points_computed, lsl_arr



from .rbc_ukqcd_params import *
from .rbc_ukqcd import *
from .jobs import *
from .params import *
from .params_meas import *
from .load_data import *
from .gen_data import *
from .auto_check import *
